from __future__ import annotations as _annotations

import asyncio
import builtins
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any, cast

import httpx
import pytest
from codex_auth_helper import (
    CodexAsyncOpenAI,
    CodexAuthConfig,
    CodexAuthStore,
    CodexOpenAI,
    CodexResponsesModel,
    CodexTokenManager,
    create_codex_async_openai,
    create_codex_chat_openai,
    create_codex_openai,
    create_codex_responses_model,
)
from langchain_openai import ChatOpenAI
from pydantic_ai.messages import ModelResponse, TextPart
from pydantic_ai.models import ModelRequestParameters
from pydantic_ai.models.openai import OpenAIResponsesModel
from typing_extensions import Sentinel

from .support import write_auth_file

_STREAM_EVENT = Sentinel("_STREAM_EVENT")


def _config(auth_path: Path) -> CodexAuthConfig:
    return CodexAuthConfig(auth_path=auth_path)


def test_create_codex_responses_model_returns_openai_responses_model(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_demo")

    model = create_codex_responses_model(
        "gpt-5",
        config=_config(auth_path),
        instructions="Answer tersely.",
    )

    assert isinstance(model, OpenAIResponsesModel)
    assert isinstance(model, CodexResponsesModel)
    assert isinstance(model.client, CodexAsyncOpenAI)
    assert str(model.client.base_url) == "https://chatgpt.com/backend-api/codex/"
    assert model.client.token_manager.current_account_id == "acct_demo"
    assert model.settings == {"openai_store": False}


def test_create_codex_responses_model_merges_settings(tmp_path: Path) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_demo")

    model = create_codex_responses_model(
        "gpt-5",
        config=_config(auth_path),
        instructions="Answer tersely.",
        settings={
            "openai_reasoning_summary": "concise",
            "openai_store": True,
        },
    )

    assert model.settings == {
        "openai_reasoning_summary": "concise",
        "openai_store": False,
    }


def test_create_codex_responses_model_rejects_missing_instructions_runtime(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_demo")

    with pytest.raises(ValueError, match="`instructions` is required"):
        create_codex_responses_model(
            "gpt-5",
            config=_config(auth_path),
            instructions=cast("Any", None),
        )


@pytest.mark.asyncio
async def test_codex_responses_model_forces_streaming_on_request(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_demo")
    model = create_codex_responses_model(
        "gpt-5",
        config=_config(auth_path),
        instructions="Answer tersely.",
    )
    expected_response = ModelResponse(parts=[TextPart("ok")], model_name="gpt-5")
    seen_stream_values: list[bool] = []

    class FakeRawResponse:
        async def __aenter__(self) -> FakeRawResponse:
            return self

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc: BaseException | None,
            traceback: Any,
        ) -> None:
            del exc_type, exc, traceback

    class FakeProcessedResponse:
        def __aiter__(self) -> Any:
            return self._iterator()

        async def _iterator(self) -> Any:
            yield _STREAM_EVENT

        def get(self) -> ModelResponse:
            return expected_response

    async def fake_responses_create(
        messages: list[Any],
        stream: bool,
        model_settings: dict[str, Any],
        model_request_parameters: ModelRequestParameters,
    ) -> FakeRawResponse:
        del messages, model_settings, model_request_parameters
        seen_stream_values.append(stream)
        return FakeRawResponse()

    async def fake_process_streamed_response(
        response: FakeRawResponse,
        model_settings: dict[str, Any],
        model_request_parameters: ModelRequestParameters,
    ) -> FakeProcessedResponse:
        del response, model_settings, model_request_parameters
        return FakeProcessedResponse()

    monkeypatch.setattr(model, "_responses_create", fake_responses_create)
    monkeypatch.setattr(model, "_process_streamed_response", fake_process_streamed_response)

    response = await model.request([], None, ModelRequestParameters())

    assert seen_stream_values == [True]
    assert response is expected_response


@pytest.mark.asyncio
async def test_codex_token_manager_refreshes_expired_tokens(tmp_path: Path) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        account_id="acct_before",
        refresh_token="refresh_before",
    )

    refreshed_access = "refreshed_access"
    refreshed_refresh = "refreshed_refresh"
    requests: list[httpx.Request] = []

    def handler(request: httpx.Request) -> httpx.Response:
        requests.append(request)
        return httpx.Response(
            status_code=200,
            json={
                "access_token": refreshed_access,
                "id_token": None,
                "refresh_token": refreshed_refresh,
            },
        )

    http_client = httpx.AsyncClient(transport=httpx.MockTransport(handler))
    token_manager = CodexTokenManager(
        config=_config(auth_path),
        store=CodexAuthStore(auth_path),
        http_client=http_client,
    )

    token = await token_manager.get_access_token()

    assert token == refreshed_access
    assert len(requests) == 1
    assert requests[0].url == httpx.URL("https://auth.openai.com/oauth/token")
    assert requests[0].headers["Content-Type"] == "application/x-www-form-urlencoded"
    assert requests[0].content.decode("utf-8") == (
        "client_id=app_EMoamEEZ73f0CkXaXp7hrann&"
        "grant_type=refresh_token&refresh_token=refresh_before"
    )

    persisted_state = CodexAuthStore(auth_path).read_state()
    assert persisted_state.access_token == refreshed_access
    assert persisted_state.refresh_token == refreshed_refresh
    assert persisted_state.account_id == "acct_before"

    await http_client.aclose()


@pytest.mark.asyncio
async def test_codex_async_openai_adds_chatgpt_account_header(tmp_path: Path) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_header")
    client = create_codex_async_openai(
        config=_config(auth_path),
        http_client=httpx.AsyncClient(),
    )

    assert client.default_headers["ChatGPT-Account-Id"] == "acct_header"

    await client.token_manager.http_client.aclose()


@pytest.mark.asyncio
async def test_codex_async_openai_covers_missing_account_header_and_owned_close(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="")
    client = create_codex_async_openai(config=_config(auth_path))

    assert "ChatGPT-Account-Id" not in client.default_headers
    assert client.default_headers["originator"] == "codex-auth-helper"
    assert client.token_manager.owns_http_client is True

    await client.close()
    assert client.token_manager.http_client.is_closed is True


@pytest.mark.asyncio
async def test_codex_async_openai_uses_codex_base_url_for_responses_requests(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_header")
    seen_urls: list[httpx.URL] = []

    def handler(request: httpx.Request) -> httpx.Response:
        seen_urls.append(request.url)
        return httpx.Response(status_code=200, json={"ok": True})

    client = create_codex_async_openai(
        config=_config(auth_path),
        http_client=httpx.AsyncClient(transport=httpx.MockTransport(handler)),
    )

    await client.responses.create(
        model="gpt-5",
        input="hello",
        store=False,
        stream=False,
    )

    assert seen_urls == [httpx.URL("https://chatgpt.com/backend-api/codex/responses")]

    await client.close()


def test_codex_auth_store_missing_file_message(tmp_path: Path) -> None:
    auth_path = tmp_path / "missing-auth.json"
    store = CodexAuthStore(auth_path)

    with pytest.raises(FileNotFoundError, match="Codex auth file was not found"):
        store.read_state()


def test_create_codex_openai_returns_sync_openai_client(tmp_path: Path) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_sync")

    client = create_codex_openai(config=_config(auth_path))

    assert isinstance(client, CodexOpenAI)
    assert client.default_headers["ChatGPT-Account-Id"] == "acct_sync"
    assert client.default_headers["originator"] == "codex-auth-helper"
    assert str(client.base_url) == "https://chatgpt.com/backend-api/codex/"

    client.close()


def test_create_codex_chat_openai_returns_langchain_chat_model(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_langchain")

    model = create_codex_chat_openai(
        "gpt-5",
        config=_config(auth_path),
        instructions="Answer tersely.",
        reasoning={"effort": "medium"},
        use_previous_response_id=True,
    )

    assert isinstance(model, ChatOpenAI)
    assert isinstance(model.root_async_client, CodexAsyncOpenAI)
    assert isinstance(model.root_client, CodexOpenAI)
    assert model.use_responses_api is True
    assert model.output_version == "responses/v1"
    assert model.use_previous_response_id is True
    assert model.reasoning == {"effort": "medium"}
    assert model.model_kwargs["instructions"] == "Answer tersely."
    assert model.store is False
    assert model.root_async_client.token_manager.current_account_id == "acct_langchain"

    model.root_client.close()
    asyncio.run(model.root_async_client.close())


def test_create_codex_chat_openai_rejects_duplicate_instructions(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_langchain")

    with pytest.raises(ValueError, match="dedicated parameter"):
        create_codex_chat_openai(
            "gpt-5",
            config=_config(auth_path),
            instructions="Answer tersely.",
            model_kwargs={"instructions": "Be concise."},
        )


def test_create_codex_chat_openai_rejects_store_override(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_langchain")

    with pytest.raises(ValueError, match="always forces `store=False`"):
        create_codex_chat_openai(
            "gpt-5",
            config=_config(auth_path),
            instructions="Answer tersely.",
            model_kwargs={"store": True},
        )


def test_create_codex_chat_openai_reports_missing_optional_dependency(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    original_import = builtins.__import__

    def fake_import(
        name: str,
        globals: dict[str, Any] | None = None,
        locals: dict[str, Any] | None = None,
        fromlist: tuple[str, ...] = (),
        level: int = 0,
    ) -> Any:
        if name == "langchain_openai":
            raise ModuleNotFoundError("No module named 'langchain_openai'")
        return original_import(name, globals, locals, fromlist, level)

    assert fake_import("math").__name__ == "math"
    monkeypatch.setattr(builtins, "__import__", fake_import)

    with pytest.raises(ModuleNotFoundError, match="codex-auth-helper\\[langchain\\]"):
        create_codex_chat_openai("gpt-5", instructions="Answer tersely.")


def test_create_codex_chat_openai_rejects_missing_instructions_runtime(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_langchain")

    with pytest.raises(ValueError, match="`instructions` is required"):
        create_codex_chat_openai(
            "gpt-5",
            config=_config(auth_path),
            instructions=cast("Any", None),
        )


@pytest.mark.asyncio
async def test_codex_openai_close_schedules_token_cleanup_in_running_loop(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_sync")
    sync_http_client = httpx.Client()
    client = create_codex_openai(
        config=_config(auth_path),
        http_client=sync_http_client,
    )

    client.close()
    await asyncio.sleep(0)

    assert sync_http_client.is_closed is False
    assert client.token_manager.http_client.is_closed is True
    sync_http_client.close()


def test_codex_openai_close_skips_async_cleanup_when_token_manager_is_not_owner(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_manual")
    async_http_client = httpx.AsyncClient()
    sync_http_client = httpx.Client()
    token_manager = CodexTokenManager(
        config=_config(auth_path),
        store=CodexAuthStore(auth_path),
        http_client=async_http_client,
        owns_http_client=False,
    )
    client = CodexOpenAI(
        base_url="https://chatgpt.com/backend-api/codex",
        http_client=sync_http_client,
        token_manager=token_manager,
        owns_http_client=False,
    )

    client.close()

    assert sync_http_client.is_closed is False
    assert async_http_client.is_closed is False
    sync_http_client.close()
    asyncio.run(async_http_client.aclose())



from __future__ import annotations as _annotations

from dataclasses import dataclass, field
from typing import Any

from acp.interfaces import Agent as AcpAgent
from acp.schema import (
    AgentMessageChunk,
    AgentPlanUpdate,
    AgentThoughtChunk,
    AllowedOutcome,
    AvailableCommandsUpdate,
    ConfigOptionUpdate,
    CreateTerminalResponse,
    CurrentModeUpdate,
    DeniedOutcome,
    EnvVariable,
    KillTerminalResponse,
    PermissionOption,
    ReadTextFileResponse,
    ReleaseTerminalResponse,
    RequestPermissionResponse,
    SessionInfoUpdate,
    TerminalOutputResponse,
    ToolCallProgress,
    ToolCallStart,
    ToolCallUpdate,
    UsageUpdate,
    UserMessageChunk,
    WaitForTerminalExitResponse,
    WriteTextFileResponse,
)

__all__ = ("RecordingACPClient", "UpdateRecord", "agent_message_texts")


@dataclass(slots=True, frozen=True, kw_only=True)
class UpdateRecord:
    session_id: str
    update: (
        UserMessageChunk
        | AgentMessageChunk
        | AgentThoughtChunk
        | ToolCallStart
        | ToolCallProgress
        | AgentPlanUpdate
        | AvailableCommandsUpdate
        | CurrentModeUpdate
        | ConfigOptionUpdate
        | SessionInfoUpdate
        | UsageUpdate
    )


@dataclass(slots=True, kw_only=True)
class RecordingACPClient:
    updates: list[UpdateRecord] = field(default_factory=list)
    permission_option_ids: list[tuple[str, list[str], ToolCallUpdate]] = field(default_factory=list)
    permission_option_names: list[tuple[str, list[str], ToolCallUpdate]] = field(
        default_factory=list,
    )
    permission_responses: list[RequestPermissionResponse] = field(default_factory=list)
    read_calls: list[tuple[str, str, int | None, int | None]] = field(default_factory=list)
    write_calls: list[tuple[str, str, str]] = field(default_factory=list)
    create_calls: list[
        tuple[
            str,
            str,
            list[str] | None,
            str | None,
            list[EnvVariable] | None,
            int | None,
        ]
    ] = field(default_factory=list)
    output_calls: list[tuple[str, str]] = field(default_factory=list)
    release_calls: list[tuple[str, str]] = field(default_factory=list)
    wait_calls: list[tuple[str, str]] = field(default_factory=list)
    kill_calls: list[tuple[str, str]] = field(default_factory=list)
    write_response: WriteTextFileResponse | None = field(default_factory=WriteTextFileResponse)
    release_response: ReleaseTerminalResponse | None = field(
        default_factory=ReleaseTerminalResponse,
    )
    kill_response: KillTerminalResponse | None = field(default_factory=KillTerminalResponse)
    wait_response: WaitForTerminalExitResponse = field(
        default_factory=lambda: WaitForTerminalExitResponse(exit_code=0),
    )
    terminal_output_response: TerminalOutputResponse = field(
        default_factory=lambda: TerminalOutputResponse(output="terminal-output", truncated=False),
    )

    def queue_permission_selected(self, option_id: str) -> None:
        self.permission_responses.append(
            RequestPermissionResponse(
                outcome=AllowedOutcome(outcome="selected", option_id=option_id),
            ),
        )

    def queue_permission_cancelled(self) -> None:
        self.permission_responses.append(
            RequestPermissionResponse(outcome=DeniedOutcome(outcome="cancelled")),
        )

    async def request_permission(
        self,
        options: list[PermissionOption],
        session_id: str,
        tool_call: ToolCallUpdate,
        **kwargs: Any,
    ) -> RequestPermissionResponse:
        del kwargs
        self.permission_option_ids.append(
            (session_id, [option.option_id for option in options], tool_call),
        )
        self.permission_option_names.append(
            (session_id, [option.name for option in options], tool_call),
        )
        if not self.permission_responses:
            raise AssertionError("unexpected permission request")
        return self.permission_responses.pop(0)

    async def session_update(
        self,
        session_id: str,
        update: (
            UserMessageChunk
            | AgentMessageChunk
            | AgentThoughtChunk
            | ToolCallStart
            | ToolCallProgress
            | AgentPlanUpdate
            | AvailableCommandsUpdate
            | CurrentModeUpdate
            | ConfigOptionUpdate
            | SessionInfoUpdate
            | UsageUpdate
        ),
        **kwargs: Any,
    ) -> None:
        del kwargs
        self.updates.append(UpdateRecord(session_id=session_id, update=update))

    async def write_text_file(
        self,
        content: str,
        path: str,
        session_id: str,
        **kwargs: Any,
    ) -> WriteTextFileResponse | None:
        del kwargs
        self.write_calls.append((session_id, path, content))
        return self.write_response

    async def read_text_file(
        self,
        path: str,
        session_id: str,
        limit: int | None = None,
        line: int | None = None,
        **kwargs: Any,
    ) -> ReadTextFileResponse:
        del kwargs
        self.read_calls.append((session_id, path, limit, line))
        return ReadTextFileResponse(content=f"file:{path}:{line}:{limit}")

    async def create_terminal(
        self,
        command: str,
        session_id: str,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[EnvVariable] | None = None,
        output_byte_limit: int | None = None,
        **kwargs: Any,
    ) -> CreateTerminalResponse:
        del kwargs
        self.create_calls.append((session_id, command, args, cwd, env, output_byte_limit))
        return CreateTerminalResponse(terminal_id="terminal-1")

    async def terminal_output(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> TerminalOutputResponse:
        del kwargs
        self.output_calls.append((session_id, terminal_id))
        return self.terminal_output_response

    async def release_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> ReleaseTerminalResponse | None:
        del kwargs
        self.release_calls.append((session_id, terminal_id))
        return self.release_response

    async def wait_for_terminal_exit(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> WaitForTerminalExitResponse:
        del kwargs
        self.wait_calls.append((session_id, terminal_id))
        return self.wait_response

    async def kill_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> KillTerminalResponse | None:
        del kwargs
        self.kill_calls.append((session_id, terminal_id))
        return self.kill_response

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        raise AssertionError(f"unexpected extension method: {method!r} {params!r}")

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        raise AssertionError(f"unexpected extension notification: {method!r} {params!r}")

    def on_connect(self, conn: AcpAgent) -> None:
        del conn


def agent_message_texts(client: RecordingACPClient) -> list[str]:
    messages: list[str] = []
    current_message_id: str | None = None
    current_text = ""
    anonymous_message_count = 0

    for record in client.updates:
        update = record.update
        if not isinstance(update, AgentMessageChunk):
            continue
        message_id = update.message_id
        if message_id is None:
            message_id = f"anonymous:{anonymous_message_count}"
            anonymous_message_count += 1
        if current_message_id != message_id:
            if current_message_id is not None:
                messages.append(current_text)
            current_message_id = message_id
            current_text = ""
        current_text += update.content.text

    if current_message_id is not None:
        messages.append(current_text)
    return messages



from __future__ import annotations as _annotations

from contextvars import ContextVar
from copy import copy
from dataclasses import dataclass, is_dataclass, replace
from typing import Any, cast

from pydantic_ai import Agent as PydanticAgent
from pydantic_ai.capabilities import CombinedCapability, Hooks

__all__ = (
    "build_capability_override_value",
    "entry_func",
    "entry_timeout",
    "entry_tool_filters",
    "hook_registry",
    "replace_hook_entry",
    "root_capability",
    "root_capability_override",
)


@dataclass(slots=True, frozen=True)
class _CapabilityOverrideValue:
    value: CombinedCapability[Any]


def build_capability_override_value(
    capability: CombinedCapability[Any],
) -> object:
    return _CapabilityOverrideValue(capability)


def root_capability(
    agent: PydanticAgent[Any, Any],
) -> CombinedCapability[Any] | None:
    capability = getattr(agent, "_root_capability", None)
    if isinstance(capability, CombinedCapability):
        return capability
    return None


def root_capability_override(
    agent: PydanticAgent[Any, Any],
) -> ContextVar[Any] | None:
    override_capability = getattr(agent, "_override_root_capability", None)
    if isinstance(override_capability, ContextVar):
        return override_capability
    return None


def hook_registry(hooks: Hooks[Any]) -> dict[str, list[Any]] | None:
    registry = getattr(hooks, "_registry", None)
    if not isinstance(registry, dict):
        return None
    normalized_registry: dict[str, list[Any]] = {}
    for key, entries in registry.items():
        if not isinstance(key, str) or not isinstance(entries, list):
            continue
        normalized_registry[key] = entries
    return normalized_registry


def entry_func(entry: object) -> Any:
    return getattr(entry, "func", None)


def entry_timeout(entry: object) -> float | None:
    timeout = getattr(entry, "timeout", None)
    if isinstance(timeout, float | int):
        return float(timeout)
    return None


def entry_tool_filters(entry: object) -> tuple[str, ...]:
    filters = getattr(entry, "tools", None)
    if isinstance(filters, frozenset):
        return tuple(sorted(filter_name for filter_name in filters if isinstance(filter_name, str)))
    return ()


def replace_hook_entry(
    entry: object,
    *,
    func: Any,
    timeout: float | None,
) -> object | None:
    if is_dataclass(entry):
        try:
            return replace(cast("Any", entry), func=func, timeout=timeout)
        except TypeError:
            return None
    try:
        wrapped_entry = copy(entry)
    except Exception:
        return None
    try:
        wrapped_entry_any = cast("Any", wrapped_entry)
        wrapped_entry_any.func = func
        wrapped_entry_any.timeout = timeout
    except Exception:
        return None
    return wrapped_entry



from __future__ import annotations as _annotations

from ._version import __version__
from .auth import CodexAuthConfig, CodexAuthState, CodexAuthStore, CodexTokenManager
from .client import (
    CodexAsyncOpenAI,
    CodexOpenAI,
    create_codex_async_openai,
    create_codex_openai,
)
from .factory import create_codex_chat_openai, create_codex_responses_model
from .model import CodexResponsesModel

__all__ = (
    "CodexAsyncOpenAI",
    "CodexAuthConfig",
    "CodexAuthState",
    "CodexAuthStore",
    "CodexOpenAI",
    "CodexResponsesModel",
    "CodexTokenManager",
    "__version__",
    "create_codex_async_openai",
    "create_codex_chat_openai",
    "create_codex_openai",
    "create_codex_responses_model",
)



from __future__ import annotations as _annotations

import base64
import json
from collections.abc import Mapping
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Final

__all__ = ("CodexAuthState",)

_AUTH_CLAIMS_KEY: Final[str] = "https://api.openai.com/auth"
_ORGANIZATIONS_KEY: Final[str] = "organizations"
JsonPrimitive = None | bool | int | float | str
JsonValue = JsonPrimitive | list["JsonValue"] | dict[str, "JsonValue"]


def _require_str(data: dict[str, JsonValue], key: str) -> str:
    value = data.get(key)
    if not isinstance(value, str) or not value:
        raise ValueError(f"Expected a non-empty string for `{key}`.")
    return value


def _optional_str(data: dict[str, JsonValue], key: str) -> str | None:
    value = data.get(key)
    return value if isinstance(value, str) and value else None


def _as_string_mapping(value: JsonValue) -> dict[str, JsonValue] | None:
    if not isinstance(value, dict):
        return None

    normalized: dict[str, JsonValue] = {}
    for key, item in value.items():
        if isinstance(key, str):
            normalized[key] = item
    return normalized


def _parse_timestamp(value: str | None) -> datetime | None:
    if value is None:
        return None
    normalized = value.replace("Z", "+00:00")
    return datetime.fromisoformat(normalized).astimezone(UTC)


def _encode_timestamp(value: datetime | None) -> str | None:
    if value is None:
        return None
    return value.astimezone(UTC).isoformat().replace("+00:00", "Z")


def _parse_jwt_claims(token: str) -> dict[str, JsonValue] | None:
    parts = token.split(".")
    if len(parts) != 3:
        return None
    payload = parts[1] + "=" * (-len(parts[1]) % 4)
    try:
        claims = json.loads(base64.urlsafe_b64decode(payload.encode("utf-8")))
    except (ValueError, json.JSONDecodeError):
        return None
    return _as_string_mapping(claims)


def _extract_account_id_from_claims(claims: Mapping[str, JsonValue]) -> str | None:
    direct_account_id = claims.get("chatgpt_account_id")
    if isinstance(direct_account_id, str) and direct_account_id:
        return direct_account_id

    auth_claims = claims.get(_AUTH_CLAIMS_KEY)
    auth_mapping = _as_string_mapping(auth_claims)
    if auth_mapping is not None:
        nested_account_id = auth_mapping.get("chatgpt_account_id")
        if isinstance(nested_account_id, str) and nested_account_id:
            return nested_account_id

    organizations = claims.get(_ORGANIZATIONS_KEY)
    if isinstance(organizations, list) and organizations:
        organization = _as_string_mapping(organizations[0])
        if organization is not None:
            organization_id = organization.get("id")
            if isinstance(organization_id, str) and organization_id:
                return organization_id

    return None


def _extract_account_id(
    *,
    access_token: str,
    account_id: str | None,
    id_token: str | None,
) -> str | None:
    if account_id is not None:
        return account_id

    if id_token is not None:
        id_claims = _parse_jwt_claims(id_token)
        if id_claims is not None:
            id_account_id = _extract_account_id_from_claims(id_claims)
            if id_account_id is not None:
                return id_account_id

    access_claims = _parse_jwt_claims(access_token)
    if access_claims is None:
        return None
    return _extract_account_id_from_claims(access_claims)


def _extract_expiry(*, access_token: str, id_token: str | None) -> datetime | None:
    for token in (id_token, access_token):
        if token is None:
            continue
        claims = _parse_jwt_claims(token)
        if claims is None:
            continue
        exp = claims.get("exp")
        if isinstance(exp, int):
            return datetime.fromtimestamp(exp, tz=UTC)
    return None


@dataclass(frozen=True, slots=True)
class CodexAuthState:
    access_token: str
    refresh_token: str
    account_id: str | None = None
    auth_mode: str | None = None
    expires_at: datetime | None = None
    id_token: str | None = None
    last_refresh: datetime | None = None
    openai_api_key: str | None = None

    @classmethod
    def from_json_dict(cls, data: dict[str, JsonValue]) -> CodexAuthState:
        tokens = _as_string_mapping(data.get("tokens"))
        if tokens is None:
            raise ValueError("Expected `tokens` to be an object.")

        access_token = _require_str(tokens, "access_token")
        id_token = _optional_str(tokens, "id_token")

        return cls(
            access_token=access_token,
            refresh_token=_require_str(tokens, "refresh_token"),
            account_id=_extract_account_id(
                access_token=access_token,
                account_id=_optional_str(tokens, "account_id"),
                id_token=id_token,
            ),
            auth_mode=_optional_str(data, "auth_mode"),
            expires_at=_extract_expiry(access_token=access_token, id_token=id_token),
            id_token=id_token,
            last_refresh=_parse_timestamp(_optional_str(data, "last_refresh")),
            openai_api_key=_optional_str(data, "OPENAI_API_KEY"),
        )

    def to_json_dict(self) -> dict[str, JsonValue]:
        return {
            "OPENAI_API_KEY": self.openai_api_key,
            "auth_mode": self.auth_mode,
            "last_refresh": _encode_timestamp(self.last_refresh),
            "tokens": {
                "access_token": self.access_token,
                "account_id": self.account_id,
                "id_token": self.id_token,
                "refresh_token": self.refresh_token,
            },
        }



from __future__ import annotations as _annotations

from collections.abc import Awaitable
from dataclasses import dataclass
from inspect import isawaitable
from typing import Any, Protocol, cast

from langgraph.graph.state import CompiledStateGraph

from .session.state import AcpSessionContext

__all__ = (
    "CompiledAgentGraph",
    "FactoryGraphSource",
    "GraphFactory",
    "GraphSource",
    "StaticGraphSource",
)

CompiledAgentGraph = CompiledStateGraph[Any, Any, Any, Any]


class GraphFactory(Protocol):
    def __call__(
        self,
        session: AcpSessionContext,
    ) -> CompiledAgentGraph | Awaitable[CompiledAgentGraph]: ...


class GraphSource(Protocol):
    async def get_graph(self, session: AcpSessionContext) -> CompiledAgentGraph: ...


@dataclass(slots=True)
class StaticGraphSource:
    graph: CompiledAgentGraph

    async def get_graph(self, session: AcpSessionContext) -> CompiledAgentGraph:
        del session
        return self.graph


@dataclass(slots=True)
class FactoryGraphSource:
    factory: GraphFactory

    async def get_graph(self, session: AcpSessionContext) -> CompiledAgentGraph:
        candidate = self.factory(session)
        if isawaitable(candidate):
            return cast("CompiledAgentGraph", await candidate)
        return cast("CompiledAgentGraph", candidate)



from __future__ import annotations as _annotations

from .fakes import RecordingACPClient, UpdateRecord, agent_message_texts
from .harness import BlackBoxHarness

__all__ = (
    "BlackBoxHarness",
    "RecordingACPClient",
    "UpdateRecord",
    "agent_message_texts",
)



from __future__ import annotations as _annotations

from types import SimpleNamespace
from typing import Any, cast

import pytest
from acp.schema import AgentMessageChunk, PermissionOption, RequestPermissionResponse
from langchain_core.messages import AIMessage

from .support import (
    GenericFakeChatModel,
    RecordingACPClient,
    agent_message_texts,
    text_block,
)


class _RunManager:
    def __init__(self) -> None:
        self.tokens: list[str] = []

    def on_llm_new_token(self, token: str, *, chunk: Any) -> None:
        del chunk
        self.tokens.append(token)


def test_generic_fake_chat_model_support_stream_paths() -> None:
    manager = _RunManager()
    model = GenericFakeChatModel(
        messages=iter([AIMessage(content="alpha,beta", id="stream-1")]),
        stream_delimiter=",",
    )
    chunks = list(model._stream([], run_manager=cast("Any", manager)))
    assert [chunk.message.content for chunk in chunks] == ["alpha", "beta"]
    assert manager.tokens == ["alpha", "beta"]

    tool_manager = _RunManager()
    tool_model = GenericFakeChatModel(
        messages=iter(
            [
                AIMessage(
                    content="",
                    id="tool-1",
                    tool_calls=[
                        {
                            "name": "demo",
                            "args": {},
                            "id": "call-1",
                            "type": "tool_call",
                        }
                    ],
                ),
            ],
        ),
    )
    tool_chunks = list(tool_model._stream([], run_manager=cast("Any", tool_manager)))
    assert len(tool_chunks) == 1
    assert tool_manager.tokens == [""]

    silent_model = GenericFakeChatModel(messages=iter([AIMessage(content="", id="silent-1")]))
    assert list(silent_model._stream([])) == []

    invalid_message_model = GenericFakeChatModel(messages=iter(["ignored"]))
    invalid_message_model._generate = cast(  # type: ignore[method-assign]
        "Any",
        lambda *args, **kwargs: SimpleNamespace(
            generations=[SimpleNamespace(message=cast("Any", object()))],
        ),
    )
    with pytest.raises(ValueError, match="Expected `AIMessage`"):
        list(invalid_message_model._stream([]))

    invalid_content_model = GenericFakeChatModel(
        messages=iter([AIMessage(content=cast("Any", ["bad-content"]), id="invalid-1")]),
    )
    with pytest.raises(ValueError, match="Expected string content"):
        list(invalid_content_model._stream([]))


@pytest.mark.asyncio
async def test_langchain_recording_acp_client_support_helpers() -> None:
    client = RecordingACPClient()
    option = PermissionOption(option_id="allow_once", name="Allow once", kind="allow_once")
    tool_call = cast("Any", object())

    with pytest.raises(AssertionError, match="unexpected permission request"):
        await client.request_permission([option], "session-1", tool_call)

    client.queue_permission_selected("allow_once")
    selected = await client.request_permission([option], "session-1", tool_call)
    assert isinstance(selected, RequestPermissionResponse)

    client.queue_permission_cancelled()
    cancelled = await client.request_permission([option], "session-1", tool_call)
    assert isinstance(cancelled, RequestPermissionResponse)

    update = AgentMessageChunk(
        session_update="agent_message_chunk",
        message_id=None,
        content=text_block("hello"),
    )
    await client.session_update("session-1", update)
    await client.session_update(
        "session-1",
        AgentMessageChunk(
            session_update="agent_message_chunk",
            message_id="message-2",
            content=text_block("world"),
        ),
    )
    client.updates.append(("session-1", object()))
    assert agent_message_texts(client) == ["hello", "world"]

    assert await client.write_text_file("body", "/tmp/demo", "session-1") == client.write_response
    read_response = await client.read_text_file("/tmp/demo", "session-1")
    assert read_response.content == "file:/tmp/demo"
    created = await client.create_terminal("echo hi", "session-1")
    assert created.terminal_id == "terminal-1"
    assert (
        await client.terminal_output("session-1", "terminal-1") == client.terminal_output_response
    )
    assert await client.release_terminal("session-1", "terminal-1") == client.release_response
    assert await client.wait_for_terminal_exit("session-1", "terminal-1") == client.wait_response
    assert await client.kill_terminal("session-1", "terminal-1") == client.kill_response

    with pytest.raises(AssertionError, match="unexpected extension method"):
        await client.ext_method("demo.echo", {"value": 1})
    with pytest.raises(AssertionError, match="unexpected extension notification"):
        await client.ext_notification("demo.note", {"value": 2})

    assert client.on_connect(cast("Any", object())) is None



from __future__ import annotations as _annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from acp.interfaces import Client as AcpClient
from acp.schema import (
    CreateTerminalResponse,
    EnvVariable,
    KillTerminalResponse,
    ReleaseTerminalResponse,
    TerminalOutputResponse,
    WaitForTerminalExitResponse,
)

from ..session.state import AcpSessionContext
from .policy import HostAccessPolicy

__all__ = (
    "ClientTerminalBackend",
    "TerminalBackend",
)


class TerminalBackend(Protocol):
    async def create_terminal(
        self,
        command: str,
        *,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[EnvVariable] | None = None,
        output_byte_limit: int | None = None,
    ) -> CreateTerminalResponse: ...

    async def terminal_output(self, terminal_id: str) -> TerminalOutputResponse: ...

    async def release_terminal(
        self,
        terminal_id: str,
    ) -> ReleaseTerminalResponse | None: ...

    async def wait_for_terminal_exit(
        self,
        terminal_id: str,
    ) -> WaitForTerminalExitResponse: ...

    async def kill_terminal(
        self,
        terminal_id: str,
    ) -> KillTerminalResponse | None: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class ClientTerminalBackend:
    client: AcpClient
    session: AcpSessionContext
    access_policy: HostAccessPolicy | None = None
    workspace_root: Path | None = None

    async def create_terminal(
        self,
        command: str,
        *,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[EnvVariable] | None = None,
        output_byte_limit: int | None = None,
    ) -> CreateTerminalResponse:
        self._enforce_command(command, args=args, cwd=cwd)
        return await self.client.create_terminal(
            command=command,
            session_id=self.session.session_id,
            args=args,
            cwd=cwd,
            env=env,
            output_byte_limit=output_byte_limit,
        )

    async def terminal_output(self, terminal_id: str) -> TerminalOutputResponse:
        return await self.client.terminal_output(
            session_id=self.session.session_id,
            terminal_id=terminal_id,
        )

    async def release_terminal(
        self,
        terminal_id: str,
    ) -> ReleaseTerminalResponse | None:
        return await self.client.release_terminal(
            session_id=self.session.session_id,
            terminal_id=terminal_id,
        )

    async def wait_for_terminal_exit(
        self,
        terminal_id: str,
    ) -> WaitForTerminalExitResponse:
        return await self.client.wait_for_terminal_exit(
            session_id=self.session.session_id,
            terminal_id=terminal_id,
        )

    async def kill_terminal(
        self,
        terminal_id: str,
    ) -> KillTerminalResponse | None:
        return await self.client.kill_terminal(
            session_id=self.session.session_id,
            terminal_id=terminal_id,
        )

    def _enforce_command(
        self,
        command: str,
        *,
        args: list[str] | None,
        cwd: str | None,
    ) -> None:
        if self.access_policy is None:
            return
        self.access_policy.enforce_command(
            command,
            args=args,
            cwd=cwd,
            session_cwd=self.session.cwd,
            workspace_root=self.workspace_root,
        )



from __future__ import annotations as _annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any, cast

import pytest
from acp import text_block, update_agent_message
from acp.exceptions import RequestError
from acp.interfaces import Agent, Client
from acp.schema import (
    AgentMessageChunk,
    AudioContentBlock,
    ClientCapabilities,
    EmbeddedResourceContentBlock,
    HttpMcpServer,
    ImageContentBlock,
    Implementation,
    InitializeResponse,
    McpServerStdio,
    NewSessionResponse,
    PermissionOption,
    PromptResponse,
    RequestPermissionResponse,
    ResourceContentBlock,
    SessionNotification,
    SseMcpServer,
    TextContentBlock,
    ToolCallUpdate,
)
from acpremote import TransportOptions, connect_remote_agent, serve_remote_agent
from acpremote.auth import bearer_headers
from acpremote.client import _merge_headers
from acpremote.metadata import TransportMetadata
from acpremote.stream import open_websocket_stream_bridge
from websockets.exceptions import ConnectionClosedOK


@dataclass(slots=True)
class _FakeWebSocket:
    incoming: asyncio.Queue[str | bytes] = field(default_factory=asyncio.Queue)
    sent: list[str] = field(default_factory=list)
    send_error: Exception | None = None
    closed: bool = False
    wait_closed_calls: int = 0

    async def recv(self, decode: bool | None = None) -> str | bytes:
        del decode
        if self.closed and self.incoming.empty():
            raise ConnectionClosedOK(None, None)
        item = await self.incoming.get()
        return item

    async def send(self, message: str) -> None:
        if self.send_error is not None:
            raise self.send_error
        self.sent.append(message)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        del code, reason
        self.closed = True

    async def wait_closed(self) -> None:
        self.wait_closed_calls += 1


@dataclass(slots=True)
class _RecordingClient:
    updates: list[SessionNotification] = field(default_factory=list)

    async def request_permission(
        self,
        options: list[PermissionOption],
        session_id: str,
        tool_call: ToolCallUpdate,
        **kwargs: Any,
    ) -> RequestPermissionResponse:
        del options, session_id, tool_call, kwargs
        raise AssertionError("permission flow is not used in phase 1")

    async def session_update(self, session_id: str, update: Any, **kwargs: Any) -> None:
        self.updates.append(
            SessionNotification(session_id=session_id, update=update, field_meta=kwargs or None),
        )

    async def write_text_file(
        self,
        content: str,
        path: str,
        session_id: str,
        **kwargs: Any,
    ) -> None:
        del content, path, session_id, kwargs
        raise AssertionError("filesystem flow is not used in phase 1")

    async def read_text_file(
        self,
        path: str,
        session_id: str,
        limit: int | None = None,
        line: int | None = None,
        **kwargs: Any,
    ) -> Any:
        del path, session_id, limit, line, kwargs
        raise AssertionError("filesystem flow is not used in phase 1")

    async def create_terminal(
        self,
        command: str,
        session_id: str,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[Any] | None = None,
        output_byte_limit: int | None = None,
        **kwargs: Any,
    ) -> Any:
        del command, session_id, args, cwd, env, output_byte_limit, kwargs
        raise AssertionError("terminal flow is not used in phase 1")

    async def terminal_output(self, session_id: str, terminal_id: str, **kwargs: Any) -> Any:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not used in phase 1")

    async def release_terminal(self, session_id: str, terminal_id: str, **kwargs: Any) -> None:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not used in phase 1")

    async def wait_for_terminal_exit(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> Any:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not used in phase 1")

    async def kill_terminal(self, session_id: str, terminal_id: str, **kwargs: Any) -> None:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not used in phase 1")

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        del params
        raise RequestError.method_not_found(method)

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        del method, params

    def on_connect(self, conn: Agent) -> None:
        del conn


@pytest.mark.asyncio
async def test_phase1_recording_client_stub_methods_cover_error_paths() -> None:
    client = _RecordingClient()

    with pytest.raises(AssertionError, match="permission flow"):
        await client.request_permission([], "session-1", cast("Any", object()))
    with pytest.raises(AssertionError, match="filesystem flow"):
        await client.write_text_file("content", "/tmp/demo", "session-1")
    with pytest.raises(AssertionError, match="filesystem flow"):
        await client.read_text_file("/tmp/demo", "session-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.create_terminal("echo hi", "session-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.terminal_output("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.release_terminal("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.wait_for_terminal_exit("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.kill_terminal("session-1", "terminal-1")

    update = AgentMessageChunk(session_update="agent_message_chunk", content=text_block("ok"))
    await client.session_update("session-1", update, source="phase1")
    assert client.updates[0].field_meta == {"source": "phase1"}
    with pytest.raises(RequestError):
        await client.ext_method("demo.missing", {})
    await client.ext_notification("demo.note", {"value": 1})
    assert client.on_connect(cast("Agent", object())) is None


@dataclass(slots=True)
class _EchoAgent:
    notifications_sent: list[str] = field(default_factory=list)
    prompts: list[str] = field(default_factory=list)
    _conn: Client | None = None

    def on_connect(self, conn: Client) -> None:
        self._conn = conn

    async def initialize(
        self,
        protocol_version: int,
        client_capabilities: ClientCapabilities | None = None,
        client_info: Implementation | None = None,
        **kwargs: Any,
    ) -> InitializeResponse:
        del client_capabilities, client_info, kwargs
        return InitializeResponse(protocol_version=protocol_version)

    async def new_session(
        self,
        cwd: str,
        mcp_servers: list[HttpMcpServer | SseMcpServer | McpServerStdio],
        **kwargs: Any,
    ) -> NewSessionResponse:
        del cwd, mcp_servers, kwargs
        return NewSessionResponse(session_id="phase1-session")

    async def prompt(
        self,
        prompt: list[
            TextContentBlock
            | ImageContentBlock
            | AudioContentBlock
            | ResourceContentBlock
            | EmbeddedResourceContentBlock
        ],
        session_id: str,
        **kwargs: Any,
    ) -> PromptResponse:
        del kwargs
        text = "".join(block.text for block in prompt if isinstance(block, TextContentBlock))
        self.prompts.append(text)
        if self._conn is not None:  # pragma: no branch
            await self._conn.session_update(
                session_id=session_id,
                update=update_agent_message(text_block(text)),
                source="acpremote-phase1",
            )
            self.notifications_sent.append(text)
        return PromptResponse(stop_reason="end_turn")


@pytest.mark.asyncio
async def test_phase1_round_trip_supports_initialize_prompt_and_session_update() -> None:
    agent = _EchoAgent()
    server = await serve_remote_agent(cast("Agent", agent))
    assert server.sockets is not None
    port = next(iter(server.sockets)).getsockname()[1]
    client = _RecordingClient()
    remote = await connect_remote_agent(cast("Client", client), f"ws://127.0.0.1:{port}/acp/ws")
    try:
        response = await remote.connection.initialize(protocol_version=1)
        assert response.protocol_version == 1
        session = await remote.connection.new_session(cwd="/tmp")
        assert session.session_id == "phase1-session"
        prompt_response = await remote.connection.prompt(
            [text_block("hello over websocket")],
            session_id=session.session_id,
        )
        assert prompt_response.stop_reason == "end_turn"
    finally:
        await remote.close()
        server.close()
        await server.wait_closed()

    assert agent.prompts == ["hello over websocket"]
    assert agent.notifications_sent == ["hello over websocket"]
    assert len(client.updates) == 1
    update = client.updates[0]
    assert update.session_id == "phase1-session"
    assert isinstance(update.update, AgentMessageChunk)
    assert update.update.content.text == "hello over websocket"
    assert update.field_meta == {"source": "acpremote-phase1"}


def test_phase1_helpers_cover_header_merge_and_transport_defaults() -> None:
    assert bearer_headers(None) is None
    assert bearer_headers("   ") is None
    assert bearer_headers("token") == {"Authorization": "Bearer token"}
    assert _merge_headers(None, {"Authorization": "Bearer token"}) == {
        "Authorization": "Bearer token",
    }
    assert _merge_headers({"X-Test": "1"}, {"Authorization": "Bearer token"}) == [
        ("X-Test", "1"),
        ("Authorization", "Bearer token"),
    ]
    assert _merge_headers([("X-Test", "1")], {"Authorization": "Bearer token"}) == [
        ("X-Test", "1"),
        ("Authorization", "Bearer token"),
    ]

    options = TransportOptions()
    assert options.compression is None
    assert options.reader_limit == 1_048_576
    assert options.max_size == 1_048_576
    assert options.max_queue == 16
    assert TransportMetadata().transport_kind == "websocket"
    assert TransportMetadata().transport_version == 1


@pytest.mark.asyncio
async def test_phase1_stream_bridge_handles_partial_lines_and_binary_frame_failures() -> None:
    websocket = _FakeWebSocket()
    await websocket.incoming.put("from remote")
    bridge = await open_websocket_stream_bridge(cast("Any", websocket))

    bridge.writer.write(b"partial")
    await bridge.writer.drain()
    assert websocket.sent == []

    bridge.writer.write(b"-line\n")
    await bridge.writer.drain()
    assert websocket.sent == ["partial-line"]
    assert await bridge.reader.readline() == b"from remote\n"
    await websocket.incoming.put(b"binary-not-supported")

    with pytest.raises(TypeError, match="binary WebSocket frames are not supported"):
        await bridge.close()

    assert websocket.closed is True
    assert websocket.wait_closed_calls == 1


@pytest.mark.asyncio
async def test_phase1_stream_bridge_supports_large_lines_with_configured_reader_limit() -> None:
    websocket = _FakeWebSocket()
    bridge = await open_websocket_stream_bridge(cast("Any", websocket), reader_limit=1_048_576)
    large_message = "x" * 70_000
    try:
        await websocket.incoming.put(large_message)
        line = await bridge.reader.readline()
    finally:
        await websocket.close()
        await websocket.incoming.put("")
        await bridge.close()

    assert line == large_message.encode("utf-8") + b"\n"


@pytest.mark.asyncio
async def test_phase1_stream_bridge_handles_sender_failures_and_transport_edges() -> None:
    websocket = _FakeWebSocket(send_error=RuntimeError("send failed"))
    bridge = await open_websocket_stream_bridge(cast("Any", websocket))

    assert bridge.writer.transport.can_write_eof() is False
    assert bridge.writer.transport.get_extra_info("missing", "fallback") == "fallback"

    bridge.writer.writelines([b"boom", b"\n"])
    with pytest.raises(RuntimeError, match="send failed"):
        await bridge.writer.drain()
    with pytest.raises(RuntimeError, match="send failed"):
        await bridge.writer.wait_closed()

    bridge.writer.transport.write_eof()
    with pytest.raises(ConnectionResetError, match="transport is closing"):
        bridge.writer.write(b"later\n")

    bridge._reader_task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await bridge._reader_task



from __future__ import annotations as _annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Any, Generic, TypeAlias, TypeVar

from acp.schema import (
    HttpMcpServer,
    ListSessionsResponse,
    LoadSessionResponse,
    McpServerStdio,
    NewSessionResponse,
    PlanEntry,
    PromptResponse,
    ResumeSessionResponse,
    SetSessionConfigOptionResponse,
    SetSessionModelResponse,
    SetSessionModeResponse,
    SseMcpServer,
    ToolCallStart,
)
from pydantic_ai import Agent as PydanticAgent
from pydantic_ai import AgentRunResult
from pydantic_ai import models as pydantic_models
from pydantic_ai.messages import ModelMessage, ToolCallPart
from pydantic_ai.output import OutputSpec
from pydantic_ai.settings import ModelSettings
from pydantic_ai.tools import DeferredToolRequests, DeferredToolResults

from .._meta_protocol import extract_field_meta
from ..approvals import ApprovalResolution
from ..awaitables import resolve_value
from ..config import AdapterConfig
from ..models import AdapterModel, ModelOverride
from ..providers import ModelSelectionState, ModeState
from ..session.state import AcpSessionContext, JsonValue, SessionTranscriptUpdate
from ._prompt_runtime import _PromptRuntime
from ._session_runtime import _SessionRuntime
from .prompts import PromptBlock, PromptInput, PromptRunOutcome
from .session_surface import ConfigOption, SessionSurface

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

RunOutputType: TypeAlias = OutputSpec[Any]

__all__ = ("_PromptRuntimeDelegationMixin", "_SessionRuntimeDelegationMixin")


class _PromptRuntimeDelegationMixin(Generic[AgentDepsT, OutputDataT]):
    """Keep prompt-runtime delegation out of the public adapter surface."""

    _config: AdapterConfig
    _prompt_runtime: _PromptRuntime[AgentDepsT, OutputDataT]

    async def _record_update(
        self,
        session: AcpSessionContext,
        update: SessionTranscriptUpdate,
    ) -> None:
        await self._prompt_runtime._record_update(session, update)

    async def _replay_transcript(self, session: AcpSessionContext) -> None:
        await self._prompt_runtime._replay_transcript(session)

    async def _run_prompt(
        self,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        prompt: list[PromptBlock],
        session: AcpSessionContext,
        output_type_override: object | None = None,
    ) -> PromptRunOutcome:
        return await self._prompt_runtime._run_prompt(
            agent=agent,
            prompt=prompt,
            session=session,
            output_type_override=output_type_override,
        )

    async def _record_tool_updates(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        messages: list[ModelMessage],
    ) -> None:
        await self._prompt_runtime._record_tool_updates(session, agent, messages)

    async def _record_bridge_updates(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        await self._prompt_runtime._record_bridge_updates(session, agent)

    async def _run_prompt_with_events(
        self,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        prompt_input: PromptInput | None,
        run_kwargs: dict[str, Any],
        session: AcpSessionContext,
    ) -> tuple[AgentRunResult[Any], bool]:
        return await self._prompt_runtime._run_prompt_with_events(
            agent=agent,
            prompt_input=prompt_input,
            run_kwargs=run_kwargs,
            session=session,
        )

    def _known_tool_call_starts(self, session: AcpSessionContext) -> dict[str, ToolCallStart]:
        return self._prompt_runtime._known_tool_call_starts(session)

    def _build_run_kwargs(
        self,
        *,
        session: AcpSessionContext,
        message_history: list[ModelMessage] | None,
        deferred_tool_results: DeferredToolResults | None,
        deps: AgentDepsT | None,
        model_override: ModelOverride | None,
        model_settings: ModelSettings | None,
        output_type: RunOutputType | None,
    ) -> dict[str, Any]:
        return self._prompt_runtime._build_run_kwargs(
            session=session,
            message_history=message_history,
            deferred_tool_results=deferred_tool_results,
            deps=deps,
            model_override=model_override,
            model_settings=model_settings,
            output_type=output_type,
        )

    def _supports_deferred_approval_bridge(self) -> bool:
        return self._prompt_runtime._supports_deferred_approval_bridge()

    async def _resolve_deferred_approvals(
        self,
        *,
        session: AcpSessionContext,
        requests: DeferredToolRequests,
    ) -> ApprovalResolution:
        return await self._prompt_runtime._resolve_deferred_approvals(
            session=session,
            requests=requests,
        )

    async def _record_cancelled_approval(
        self,
        session: AcpSessionContext,
        tool_call: ToolCallPart | None,
    ) -> None:
        await self._prompt_runtime._record_cancelled_approval(session, tool_call)

    def _build_run_output_type(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        session: AcpSessionContext,
    ) -> RunOutputType | None:
        return self._prompt_runtime._build_run_output_type(agent, session=session)

    def _should_stream_text_responses(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        model_override: ModelOverride | None,
        output_type: RunOutputType | None,
    ) -> bool:
        return self._prompt_runtime._should_stream_text_responses(
            agent,
            model_override=model_override,
            output_type=output_type,
        )

    def _contains_text_output(self, output_type: Any) -> bool:
        return self._prompt_runtime._contains_text_output(output_type)

    def _contains_native_plan_generation(self, output_type: Any) -> bool:
        return self._prompt_runtime._contains_native_plan_generation(output_type)

    def _supports_streaming_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        model_override: ModelOverride | None,
    ) -> bool:
        return self._prompt_runtime._supports_streaming_model(
            agent,
            model_override=model_override,
        )

    def _resolve_runtime_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        model_override: ModelOverride | None,
    ) -> pydantic_models.Model:
        return self._prompt_runtime._resolve_runtime_model(
            agent,
            model_override=model_override,
        )

    def _text_chunk_from_event(
        self,
        event: Any,
    ) -> str | None:
        return self._prompt_runtime._text_chunk_from_event(event)

    def _synchronize_native_plan_output(
        self,
        session: AcpSessionContext,
        output: Any,
        *,
        streamed_output: bool,
    ) -> str:
        return self._prompt_runtime._synchronize_native_plan_output(
            session,
            output,
            streamed_output=streamed_output,
        )

    async def _resolve_prompt_model_override(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        prompt: Sequence[PromptBlock],
        model_override: ModelOverride | None,
    ) -> ModelOverride | None:
        provider = self._config.prompt_model_override_provider
        if provider is None:
            return model_override
        override = await resolve_value(
            provider.get_prompt_model_override(
                session,
                agent,
                prompt,
                model_override,
            ),
        )
        if override is None:
            return model_override
        return override


class _SessionRuntimeDelegationMixin(Generic[AgentDepsT, OutputDataT]):
    """Keep session/runtime plumbing out of the main adapter class body."""

    _adapter_prompt: Any
    _session_runtime: _SessionRuntime[AgentDepsT, OutputDataT]

    def _normalize_cwd(self, cwd: str) -> Path:
        return self._session_runtime._normalize_cwd(cwd)

    def _configure_agent_runtime(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        self._session_runtime._configure_agent_runtime(session, agent)

    def _set_active_session(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
    ) -> None:
        self._session_runtime._set_active_session(agent, session)

    def _bind_session_client(self, session: AcpSessionContext) -> AcpSessionContext:
        return self._session_runtime._bind_session_client(session)

    async def _build_session_surface(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> SessionSurface:
        return await self._session_runtime._build_session_surface(session, agent)

    async def _emit_session_state_updates(
        self,
        session: AcpSessionContext,
        surface: SessionSurface,
        *,
        emit_available_commands: bool,
        emit_config_options: bool,
        emit_current_mode: bool,
        emit_plan: bool,
        emit_session_info: bool,
    ) -> None:
        await self._session_runtime._emit_session_state_updates(
            session,
            surface,
            emit_available_commands=emit_available_commands,
            emit_config_options=emit_config_options,
            emit_current_mode=emit_current_mode,
            emit_plan=emit_plan,
            emit_session_info=emit_session_info,
        )

    async def _build_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        model_selection_state: ModelSelectionState | None,
        mode_state: ModeState | None,
    ) -> list[ConfigOption] | None:
        return await self._session_runtime._build_config_options(
            session,
            agent,
            model_selection_state=model_selection_state,
            mode_state=mode_state,
        )

    async def _get_model_selection_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModelSelectionState | None:
        return await self._session_runtime._get_model_selection_state(session, agent)

    async def _set_provider_model_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        model_id: str,
    ) -> ModelSelectionState | None:
        return await self._session_runtime._set_provider_model_state(session, agent, model_id)

    def _synchronize_session_model_selection(
        self,
        session: AcpSessionContext,
        model_state: ModelSelectionState | None,
    ) -> None:
        self._session_runtime._synchronize_session_model_selection(session, model_state)

    async def _get_mode_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModeState | None:
        return await self._session_runtime._get_mode_state(session, agent)

    async def _set_provider_mode_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        mode_id: str,
    ) -> ModeState | None:
        return await self._session_runtime._set_provider_mode_state(session, agent, mode_id)

    def _synchronize_mode_state(
        self,
        session: AcpSessionContext,
        mode_state: ModeState | None,
    ) -> None:
        self._session_runtime._synchronize_mode_state(session, mode_state)

    async def _get_provider_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> list[ConfigOption] | None:
        return await self._session_runtime._get_provider_config_options(session, agent)

    async def _set_provider_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        config_id: str,
        value: str | bool,
    ) -> bool:
        return await self._session_runtime._set_provider_config_options(
            session,
            agent,
            config_id,
            value,
        )

    async def _get_plan_entries(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> list[PlanEntry] | None:
        return await self._session_runtime._get_plan_entries(session, agent)

    async def _synchronize_session_metadata(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        await self._session_runtime._synchronize_session_metadata(session, agent)

    async def _get_approval_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> dict[str, JsonValue] | None:
        return await self._session_runtime._get_approval_state(session, agent)

    def _find_model_option(
        self,
        model_id: str,
        *,
        available_models: Sequence[AdapterModel] | None = None,
    ) -> AdapterModel | None:
        return self._session_runtime._find_model_option(
            model_id,
            available_models=available_models,
        )

    def _require_model_option(self, model_id: str) -> AdapterModel:
        return self._session_runtime._require_model_option(model_id)

    def _resolve_current_model_id(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> str | None:
        return self._session_runtime._resolve_current_model_id(session, agent)

    def _resolve_model_id_from_value(self, model_value: ModelOverride | None) -> str | None:
        return self._session_runtime._resolve_model_id_from_value(model_value)

    async def _resolve_model_override(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModelOverride | None:
        return await self._session_runtime._resolve_model_override(session, agent)

    def _supports_fallback_model_selection(self) -> bool:
        return self._session_runtime._supports_fallback_model_selection()

    async def new_session(
        self,
        cwd: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
        **kwargs: Any,
    ) -> NewSessionResponse:
        """Create a new ACP session rooted at the provided working directory."""
        del kwargs
        return await self._session_runtime.new_session(cwd, mcp_servers)

    async def load_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
        **kwargs: Any,
    ) -> LoadSessionResponse | None:
        """Load a persisted ACP session and rebuild its visible surface."""
        del kwargs
        response = await self._session_runtime.load_session(cwd, session_id, mcp_servers)
        if response is None:
            return None
        return LoadSessionResponse(
            config_options=response.config_options,
            models=response.models,
            modes=response.modes,
        )

    async def list_sessions(
        self,
        cursor: str | None = None,
        cwd: str | None = None,
        **kwargs: Any,
    ) -> ListSessionsResponse:
        """List persisted sessions visible from the current runtime."""
        del cursor, kwargs
        return await self._session_runtime.list_sessions(cwd=cwd)

    async def resume_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
        **kwargs: Any,
    ) -> ResumeSessionResponse:
        """Resume a session and replay the ACP-visible session surface."""
        del kwargs
        return await self._session_runtime.resume_session(cwd, session_id, mcp_servers)

    async def set_session_mode(
        self,
        mode_id: str,
        session_id: str,
        **kwargs: Any,
    ) -> SetSessionModeResponse | None:
        """Set the current ACP session mode through the active mode provider."""
        del kwargs
        return await self._session_runtime.set_session_mode(mode_id, session_id)

    async def set_session_model(
        self,
        model_id: str,
        session_id: str,
        **kwargs: Any,
    ) -> SetSessionModelResponse | None:
        """Set the current ACP session model selection."""
        del kwargs
        return await self._session_runtime.set_session_model(model_id, session_id)

    async def set_config_option(
        self,
        config_id: str,
        session_id: str,
        value: str | bool,
        **kwargs: Any,
    ) -> SetSessionConfigOptionResponse | None:
        """Apply a session-local config option exposed by the adapter or a bridge."""
        del kwargs
        return await self._session_runtime.set_config_option(config_id, session_id, value)

    async def prompt(
        self,
        prompt: list[PromptBlock],
        session_id: str,
        message_id: str | None = None,
        **kwargs: Any,
    ) -> PromptResponse:
        """Run one ACP prompt turn against the bound Pydantic AI agent."""
        return await self._adapter_prompt.prompt(
            prompt,
            session_id,
            message_id=message_id,
            request_meta=extract_field_meta(kwargs),
        )



from __future__ import annotations as _annotations

import json
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Protocol

from acp.schema import (
    ContentToolCallContent,
    FileEditToolCallContent,
    TerminalToolCallContent,
    TextContentBlock,
    ToolCallLocation,
    ToolCallProgress,
    ToolCallStart,
    ToolCallStatus,
    ToolKind,
)
from pydantic_ai import (
    ModelMessage,
    ModelResponse,
    RetryPromptPart,
    ToolCallPart,
    ToolReturnPart,
)
from pydantic_ai.messages import (
    CompactionPart,
    NativeToolCallPart,
    NativeToolReturnPart,
)
from typing_extensions import TypeIs

from ._projection_text import format_code_block, single_line_summary, truncate_text
from .serialization import OutputSerializer
from .session.state import JsonValue

__all__ = (
    "BuiltinToolProjectionMap",
    "CompositeProjectionMap",
    "DefaultToolClassifier",
    "FileSystemProjectionMap",
    "HarnessCodeModeProjectionMap",
    "HarnessFileSystemProjectionMap",
    "HarnessShellProjectionMap",
    "ProjectionAwareToolClassifier",
    "ProjectionMap",
    "ToolClassifier",
    "WebToolProjectionMap",
    "build_compaction_updates",
    "build_tool_progress_update",
    "build_tool_start_update",
    "build_tool_updates",
    "compose_projection_maps",
    "extract_tool_call_locations",
)

_PATH_KEYS = (
    "destination_path",
    "file_path",
    "filepath",
    "path",
    "source_path",
    "target_path",
)
_CONTENT_KEYS = ("content", "text", "new_text")
_OLD_TEXT_KEYS = ("old_text", "oldText", "previous_content", "previous_text")
_COMMAND_KEYS = ("command", "cmd", "script", "bash")
_SEARCH_PATTERN_KEYS = ("pattern", "query", "q", "glob", "search")
_TERMINAL_ID_KEYS = ("terminal_id", "terminalId")
_MAX_COMMAND_PREVIEW_CHARS = 4000
_MAX_COMMAND_TITLE_CHARS = 80
_MAX_FILE_READ_PREVIEW_CHARS = 12000
_MAX_WEB_PREVIEW_CHARS = 2000
_DEFAULT_SEARCH_TOOL_NAMES = frozenset(
    {"duckduckgo_search", "exa_search", "tavily_search", "web_search"},
)
_DEFAULT_FETCH_TOOL_NAMES = frozenset({"web_fetch"})
_DEFAULT_IMAGE_GENERATION_TOOL_NAMES = frozenset({"generate_image", "image_generation"})
_DEFAULT_MCP_TOOL_NAME_PREFIXES = frozenset({"mcp_server:"})
_HARNESS_READ_TOOL_NAMES = frozenset({"read_file"})
_HARNESS_WRITE_TOOL_NAMES = frozenset({"write_file", "create_directory"})
_HARNESS_EDIT_TOOL_NAMES = frozenset({"edit_file"})
_HARNESS_SEARCH_TOOL_NAMES = frozenset({"list_directory", "search_files"})
_HARNESS_SHELL_TOOL_NAMES = frozenset(
    {"run_command", "start_command", "check_command", "stop_command"},
)
_HARNESS_CODE_MODE_TOOL_NAMES = frozenset({"run_code"})


def _is_string_keyed_object_dict(value: Any) -> TypeIs[dict[str, Any]]:
    return isinstance(value, dict) and all(isinstance(key, str) for key in value)


class ToolClassifier(Protocol):
    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind: ...

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class ToolProjection:
    content: (
        list[ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent] | None
    ) = None
    locations: list[ToolCallLocation] | None = None
    status: ToolCallStatus | None = None
    title: str | None = None


class ProjectionMap(Protocol):
    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None: ...

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class CompositeProjectionMap:
    maps: tuple[ProjectionMap, ...]

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        return _merge_tool_projections(
            projection_map.project_start(tool_name, cwd=cwd, raw_input=raw_input)
            for projection_map in self.maps
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        return _merge_tool_projections(
            projection_map.project_progress(
                tool_name,
                cwd=cwd,
                raw_input=raw_input,
                raw_output=raw_output,
                serialized_output=serialized_output,
                status=status,
            )
            for projection_map in self.maps
        )


def compose_projection_maps(
    projection_maps: Sequence[ProjectionMap] | None,
) -> ProjectionMap | None:
    if projection_maps is None:
        return None
    if len(projection_maps) == 0:
        return None
    if len(projection_maps) == 1:
        return projection_maps[0]
    return CompositeProjectionMap(maps=tuple(projection_maps))


@dataclass(slots=True, frozen=True, kw_only=True)
class FileSystemProjectionMap:
    write_tool_names: frozenset[str] = frozenset()
    read_tool_names: frozenset[str] = frozenset()
    bash_tool_names: frozenset[str] = frozenset()
    search_tool_names: frozenset[str] = frozenset()
    default_write_tool: str | None = None
    default_read_tool: str | None = None
    default_bash_tool: str | None = None
    default_search_tool: str | None = None
    path_arg: str | None = None
    content_arg: str | None = None
    old_text_arg: str | None = None
    command_arg: str | None = None
    terminal_id_arg: str | None = None
    search_path_arg: str | None = None
    search_pattern_arg: str | None = None
    render_search_results_as_tree: bool = False
    hide_dot_directories_in_tree: bool = True
    tree_root_label: str | None = None

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        if tool_name in self._search_tool_names():
            if not _is_string_keyed_object_dict(raw_input):
                return None
            search_text = self._format_search_start(raw_input)
            return ToolProjection(
                content=[ContentToolCallContent(type="content", content=_text_block(search_text))],
                locations=self._search_locations_from_input(raw_input),
                title=self._format_search_title(raw_input),
            )
        if tool_name in self._bash_tool_names():
            if not _is_string_keyed_object_dict(raw_input):
                return None
            command = self._command_from_input(raw_input)
            if command is None:
                return None
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(_format_command_preview(command)),
                    ),
                ],
                title=_format_command_title(command),
            )
        if tool_name not in self._write_tool_names():
            return None
        if not _is_string_keyed_object_dict(raw_input):
            return None
        path = self._path_from_input(raw_input)
        new_text = self._content_from_input(raw_input)
        if path is None or new_text is None:
            return None
        return ToolProjection(
            content=[
                FileEditToolCallContent(
                    type="diff",
                    path=path,
                    old_text=self._old_text_from_input(raw_input, path=path, cwd=cwd),
                    new_text=new_text,
                ),
            ],
            locations=[ToolCallLocation(path=path)],
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        if tool_name in self._bash_tool_names():
            terminal_id = self._terminal_id_from_value(raw_output)
            status_override = _bash_status(raw_output, fallback=status)
            if terminal_id is not None:
                return ToolProjection(
                    content=[TerminalToolCallContent(type="terminal", terminal_id=terminal_id)],
                    status=status_override,
                )
            return ToolProjection(
                content=_bash_progress_content(
                    raw_input=raw_input,
                    raw_output=raw_output,
                    serialized_output=serialized_output,
                ),
                status=status_override,
            )
        if tool_name in self._search_tool_names():
            if not _is_string_keyed_object_dict(raw_input):
                return None
            output_text = _stringify_value(raw_output, serialized_output)
            if (
                status == "completed"
                and self.render_search_results_as_tree
                and not _looks_like_status_or_error_output(output_text)
            ):
                rendered_output = _render_path_tree(
                    output_text,
                    root_label=self._tree_root_label(raw_input),
                    hide_dot_directories=self.hide_dot_directories_in_tree,
                )
            else:
                rendered_output = output_text
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(rendered_output),
                    ),
                ],
                locations=self._search_locations_from_input(raw_input),
                status=status,
            )
        if status != "completed":
            return None
        if tool_name in self._write_tool_names():
            return self.project_start(tool_name, cwd=cwd, raw_input=raw_input)
        if tool_name not in self._read_tool_names():
            return None
        if not _is_string_keyed_object_dict(raw_input):
            return None
        path = self._path_from_input(raw_input)
        if path is None:
            return None
        new_text = (
            serialized_output
            if raw_output is None
            else _stringify_value(raw_output, serialized_output)
        )
        return ToolProjection(
            content=[
                FileEditToolCallContent(
                    type="diff",
                    path=path,
                    old_text="",
                    new_text=new_text,
                ),
            ],
            locations=[ToolCallLocation(path=path)],
        )

    def _write_tool_names(self) -> frozenset[str]:
        names = set(self.write_tool_names)
        if self.default_write_tool is not None:
            names.add(self.default_write_tool)
        return frozenset(names)

    def _read_tool_names(self) -> frozenset[str]:
        names = set(self.read_tool_names)
        if self.default_read_tool is not None:
            names.add(self.default_read_tool)
        return frozenset(names)

    def _bash_tool_names(self) -> frozenset[str]:
        names = set(self.bash_tool_names)
        if self.default_bash_tool is not None:
            names.add(self.default_bash_tool)
        return frozenset(names)

    def _search_tool_names(self) -> frozenset[str]:
        names = set(self.search_tool_names)
        if self.default_search_tool is not None:
            names.add(self.default_search_tool)
        return frozenset(names)

    def _path_from_input(self, raw_input: dict[str, Any]) -> str | None:
        for key in _candidate_keys(self.path_arg, _PATH_KEYS):
            value = raw_input.get(key)
            if isinstance(value, str) and value:
                return value
        return None

    def _content_from_input(self, raw_input: dict[str, Any]) -> str | None:
        for key in _candidate_keys(self.content_arg, _CONTENT_KEYS):
            value = raw_input.get(key)
            if value is not None:
                return _stringify_value(value, None)
        return None

    def _command_from_input(self, raw_input: dict[str, Any]) -> str | None:
        for key in _candidate_keys(self.command_arg, _COMMAND_KEYS):
            value = raw_input.get(key)
            if isinstance(value, str) and value:
                return value
        return None

    def _search_path_from_input(self, raw_input: dict[str, Any]) -> str | None:
        for key in _candidate_keys(self.search_path_arg, _PATH_KEYS):
            value = raw_input.get(key)
            if isinstance(value, str) and value:
                return value
        return None

    def _search_pattern_from_input(self, raw_input: dict[str, Any]) -> str | None:
        for key in _candidate_keys(self.search_pattern_arg, _SEARCH_PATTERN_KEYS):
            value = raw_input.get(key)
            if isinstance(value, str) and value:
                return value
        return None

    def _search_locations_from_input(
        self,
        raw_input: dict[str, Any],
    ) -> list[ToolCallLocation] | None:
        path = self._search_path_from_input(raw_input)
        return [ToolCallLocation(path=path)] if path is not None else None

    def _format_search_title(self, raw_input: dict[str, Any]) -> str:
        pattern = self._search_pattern_from_input(raw_input)
        path = self._search_path_from_input(raw_input)
        if pattern is not None and path is not None:
            return (
                f"Search {path} for {_single_line_preview(pattern, limit=_MAX_COMMAND_TITLE_CHARS)}"
            )
        if pattern is not None:
            return f"Search for {_single_line_preview(pattern, limit=_MAX_COMMAND_TITLE_CHARS)}"
        if path is not None:
            return f"List {path}"
        return "Search files"

    def _format_search_start(self, raw_input: dict[str, Any]) -> str:
        lines: list[str] = []
        path = self._search_path_from_input(raw_input)
        pattern = self._search_pattern_from_input(raw_input)
        if path is not None:
            lines.append(f"Path: {path}")
        if pattern is not None:
            lines.append(f"Pattern: {pattern}")
        return "\n".join(lines) if lines else "Searching files."

    def _tree_root_label(self, raw_input: dict[str, Any]) -> str:
        if self.tree_root_label is not None:
            return self.tree_root_label
        return self._search_path_from_input(raw_input) or "."

    def _old_text_from_input(
        self,
        raw_input: dict[str, Any],
        *,
        path: str,
        cwd: Path | None,
    ) -> str:
        for key in _candidate_keys(self.old_text_arg, _OLD_TEXT_KEYS):
            value = raw_input.get(key)
            if value is not None:
                return _stringify_value(value, None)
        return _read_existing_text(path, cwd=cwd)

    def _terminal_id_from_value(self, value: Any) -> str | None:
        if not _is_string_keyed_object_dict(value):
            return None
        for key in _candidate_keys(self.terminal_id_arg, _TERMINAL_ID_KEYS):
            terminal_id = value.get(key)
            if isinstance(terminal_id, str) and terminal_id:
                return terminal_id
        return None


def _iter_filesystem_projection_maps(
    projection_map: ProjectionMap,
) -> Iterable[FileSystemProjectionMap]:
    if isinstance(projection_map, FileSystemProjectionMap):
        yield projection_map
        return
    filesystem_projection = getattr(projection_map, "_filesystem_projection", None)
    if isinstance(filesystem_projection, FileSystemProjectionMap):
        yield filesystem_projection
        return
    if isinstance(projection_map, CompositeProjectionMap):
        for nested_map in projection_map.maps:
            yield from _iter_filesystem_projection_maps(nested_map)


@dataclass(slots=True, frozen=True, kw_only=True)
class WebToolProjectionMap:
    search_tool_names: frozenset[str] = _DEFAULT_SEARCH_TOOL_NAMES
    fetch_tool_names: frozenset[str] = _DEFAULT_FETCH_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if not _is_string_keyed_object_dict(raw_input):
            return None
        if tool_name in self.search_tool_names:
            query = _web_search_query(raw_input)
            if query is None:
                return None
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(_format_web_search_start(raw_input)),
                    ),
                ],
                title=f"Search web for {_single_line_preview(query, limit=_MAX_COMMAND_TITLE_CHARS)}",
            )
        if tool_name not in self.fetch_tool_names:
            return None
        url = _web_fetch_url(raw_input)
        if url is None:
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_fetch_start(raw_input)),
                ),
            ],
            title=f"Fetch {_single_line_preview(url, limit=_MAX_COMMAND_TITLE_CHARS)}",
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd, raw_input
        if status != "completed":
            return None
        if tool_name in self.search_tool_names:
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _format_web_search_progress(raw_output, serialized_output),
                        ),
                    ),
                ],
            )
        if tool_name not in self.fetch_tool_names:
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_fetch_progress(raw_output, serialized_output)),
                ),
            ],
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class BuiltinToolProjectionMap:
    web_projection_map: WebToolProjectionMap = field(default_factory=WebToolProjectionMap)
    image_generation_tool_names: frozenset[str] = _DEFAULT_IMAGE_GENERATION_TOOL_NAMES
    mcp_tool_name_prefixes: frozenset[str] = _DEFAULT_MCP_TOOL_NAME_PREFIXES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        web_projection = self.web_projection_map.project_start(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
        )
        if web_projection is not None:
            return web_projection
        if not _is_string_keyed_object_dict(raw_input):
            return None
        if tool_name in self.image_generation_tool_names:
            prompt = _first_string(raw_input, "prompt", "input", "text")
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(_format_image_generation_start(raw_input)),
                    ),
                ],
                title=(
                    f"Generate image for {_single_line_preview(prompt, limit=_MAX_COMMAND_TITLE_CHARS)}"
                    if prompt is not None
                    else "Generate image"
                ),
            )
        if not any(tool_name.startswith(prefix) for prefix in self.mcp_tool_name_prefixes):
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_mcp_start(tool_name, raw_input)),
                ),
            ],
            title=_format_mcp_title(tool_name, raw_input),
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        web_projection = self.web_projection_map.project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )
        if web_projection is not None:
            return web_projection
        if status != "completed":
            return None
        if tool_name in self.image_generation_tool_names:
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _format_image_generation_progress(raw_output, serialized_output),
                        ),
                    ),
                ],
            )
        if not any(tool_name.startswith(prefix) for prefix in self.mcp_tool_name_prefixes):
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_mcp_progress(raw_output, serialized_output)),
                ),
            ],
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class HarnessFileSystemProjectionMap:
    read_tool_names: frozenset[str] = _HARNESS_READ_TOOL_NAMES
    write_tool_names: frozenset[str] = _HARNESS_WRITE_TOOL_NAMES
    edit_tool_names: frozenset[str] = _HARNESS_EDIT_TOOL_NAMES
    search_tool_names: frozenset[str] = _HARNESS_SEARCH_TOOL_NAMES

    @property
    def _filesystem_projection(self) -> FileSystemProjectionMap:
        return FileSystemProjectionMap(
            read_tool_names=self.read_tool_names,
            write_tool_names=self.write_tool_names | self.edit_tool_names,
            search_tool_names=self.search_tool_names,
            path_arg="path",
            content_arg="content",
            old_text_arg="old_text",
            search_pattern_arg="pattern",
            search_path_arg="path",
            render_search_results_as_tree=True,
        )

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        if tool_name in self.read_tool_names:
            return self._project_read_start(raw_input)
        return self._filesystem_projection.project_start(tool_name, cwd=cwd, raw_input=raw_input)

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        if tool_name in self.read_tool_names:
            return self._project_read_progress(
                raw_input=raw_input,
                raw_output=raw_output,
                serialized_output=serialized_output,
                status=status,
            )
        return self._filesystem_projection.project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )

    def _project_read_start(self, raw_input: Any) -> ToolProjection | None:
        if not _is_string_keyed_object_dict(raw_input):
            return None
        path = self._read_path_from_input(raw_input)
        if path is None:
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(f"Read file: `{path}`"),
                ),
            ],
            locations=[ToolCallLocation(path=path)],
            title=f"Read {_single_line_preview(path, limit=_MAX_COMMAND_TITLE_CHARS)}",
        )

    def _project_read_progress(
        self,
        *,
        raw_input: Any,
        raw_output: Any,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        if not _is_string_keyed_object_dict(raw_input):
            return None
        path = self._read_path_from_input(raw_input)
        if path is None:
            return None
        output_text = _stringify_value(raw_output, serialized_output)
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_harness_read_progress(path, output_text)),
                ),
            ],
            locations=[ToolCallLocation(path=path)],
            status=status,
        )

    def _read_path_from_input(self, raw_input: dict[str, Any]) -> str | None:
        return self._filesystem_projection._path_from_input(raw_input)


@dataclass(slots=True, frozen=True, kw_only=True)
class HarnessShellProjectionMap:
    tool_names: frozenset[str] = _HARNESS_SHELL_TOOL_NAMES

    @property
    def _filesystem_projection(self) -> FileSystemProjectionMap:
        return FileSystemProjectionMap(
            bash_tool_names=self.tool_names,
            command_arg="command",
            terminal_id_arg="command_id",
        )

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        if tool_name in {"run_command", "start_command"}:
            return self._filesystem_projection.project_start(
                tool_name,
                cwd=cwd,
                raw_input=raw_input,
            )
        if tool_name not in self.tool_names or not _is_string_keyed_object_dict(raw_input):
            return None
        command_id = _first_string(raw_input, "command_id", "terminal_id", "terminalId")
        action = "Check" if tool_name == "check_command" else "Stop"
        if command_id is None:
            return ToolProjection(title=f"{action} command")
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(f"Command ID: {command_id}"),
                ),
            ],
            title=f"{action} command {command_id}",
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        return self._filesystem_projection.project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class HarnessCodeModeProjectionMap:
    tool_names: frozenset[str] = _HARNESS_CODE_MODE_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if tool_name not in self.tool_names or not _is_string_keyed_object_dict(raw_input):
            return None
        code = _first_string(raw_input, "code", "script", "python")
        if code is None:
            return ToolProjection(title="Run code")
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(format_code_block(code, language="python")),
                ),
            ],
            title=f"Run code: {_single_line_preview(code, limit=_MAX_COMMAND_TITLE_CHARS)}",
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd, raw_input
        if tool_name not in self.tool_names or status != "completed":
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_stringify_value(raw_output, serialized_output)),
                ),
            ],
        )


class DefaultToolClassifier:
    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind:
        del raw_input
        lowered = tool_name.lower()
        if lowered in _DEFAULT_SEARCH_TOOL_NAMES:
            return "search"
        if lowered in _DEFAULT_FETCH_TOOL_NAMES:
            return "fetch"
        if lowered in _DEFAULT_IMAGE_GENERATION_TOOL_NAMES:
            return "execute"
        if any(lowered.startswith(prefix) for prefix in _DEFAULT_MCP_TOOL_NAME_PREFIXES):
            return "execute"
        if lowered.startswith(("read_", "load_", "open_", "cat_")):
            return "read"
        if lowered.startswith(("write_", "edit_", "patch_", "update_")):
            return "edit"
        if lowered.startswith(("delete_", "remove_")):
            return "delete"
        if lowered.startswith(("move_", "rename_")):
            return "move"
        if lowered.startswith(("search_", "grep_", "find_")):
            return "search"
        if lowered.startswith(("fetch_", "scrape_", "download_")):
            return "fetch"
        if lowered.startswith(("think_", "plan_")):
            return "think"
        return "execute"

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str:
        del raw_input
        return tool_name


@dataclass(slots=True, frozen=True, kw_only=True)
class ProjectionAwareToolClassifier:
    base_classifier: ToolClassifier
    projection_maps: Sequence[ProjectionMap]

    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind:
        projection_kind = self._projection_kind(tool_name)
        if projection_kind is not None:
            return projection_kind
        return self.base_classifier.classify(tool_name, raw_input)

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str:
        return self.base_classifier.approval_policy_key(tool_name, raw_input)

    def _projection_kind(self, tool_name: str) -> ToolKind | None:
        for projection_map in self.projection_maps:
            if (
                isinstance(projection_map, HarnessCodeModeProjectionMap)
                and tool_name in projection_map.tool_names
            ):
                return "execute"
            for filesystem_map in _iter_filesystem_projection_maps(projection_map):
                if tool_name in filesystem_map._read_tool_names():
                    return "read"
                if tool_name in filesystem_map._write_tool_names():
                    return "edit"
                if tool_name in filesystem_map._bash_tool_names():
                    return "execute"
                if tool_name in filesystem_map._search_tool_names():
                    return "search"
        return None


def _is_output_tool(tool_name: str) -> bool:
    return tool_name == "final_result"


def extract_tool_call_locations(raw_input: Any) -> list[ToolCallLocation] | None:
    if not _is_string_keyed_object_dict(raw_input):
        return None

    for key in _PATH_KEYS:
        value = raw_input.get(key)
        if isinstance(value, str) and value:
            return [ToolCallLocation(path=value)]
    return None


@dataclass(slots=True)
class _PathTreeNode:
    directories: dict[str, _PathTreeNode] = field(default_factory=dict)
    files: set[str] = field(default_factory=set)


def _looks_like_status_or_error_output(text: str) -> bool:
    stripped = text.strip().lower()
    if not stripped:
        return True
    return stripped.startswith(("error:", "failed:", "status:", "traceback"))


def _render_path_tree(
    text: str,
    *,
    root_label: str,
    hide_dot_directories: bool,
) -> str:
    root = _PathTreeNode()
    has_path = False
    truncated = False
    for raw_line in text.splitlines():
        line = raw_line.strip()
        if not line:
            continue
        if line == "...":
            truncated = True
            continue
        normalized = line.removeprefix("./")
        if not normalized:
            continue
        if _path_has_hidden_directory(normalized, hide_dot_directories=hide_dot_directories):
            continue
        _insert_tree_path(root, normalized)
        has_path = True
    if not has_path:
        return text
    lines = [f"Tree: {root_label}", ""]
    entries = _render_tree_node(root, prefix="")
    lines.extend(entries)
    if truncated:
        lines.append("...")
    return "\n".join(lines).rstrip()


def _path_has_hidden_directory(path: str, *, hide_dot_directories: bool) -> bool:
    if not hide_dot_directories:
        return False
    parts = [part for part in path.strip("/").split("/") if part]
    directory_parts = parts[:-1] if not path.endswith("/") else parts
    return any(part.startswith(".") for part in directory_parts)


def _insert_tree_path(root: _PathTreeNode, path: str) -> None:
    is_directory = path.endswith("/")
    parts = [part for part in path.strip("/").split("/") if part]
    if not parts:
        return
    current = root
    for directory in parts[:-1]:
        current = current.directories.setdefault(directory, _PathTreeNode())
    leaf = parts[-1]
    if is_directory:
        current.directories.setdefault(leaf, _PathTreeNode())
    else:
        current.files.add(leaf)


def _render_tree_node(node: _PathTreeNode, *, prefix: str) -> list[str]:
    rendered: list[str] = []
    entries: list[tuple[str, str]] = [("directory", name) for name in sorted(node.directories)] + [
        ("file", name) for name in sorted(node.files)
    ]
    for index, (entry_type, name) in enumerate(entries):
        is_last = index == len(entries) - 1
        connector = "└── " if is_last else "├── "
        child_prefix = "    " if is_last else "│   "
        if entry_type == "directory":
            rendered.append(f"{prefix}{connector}{name}/")
            rendered.extend(
                _render_tree_node(
                    node.directories[name],
                    prefix=f"{prefix}{child_prefix}",
                ),
            )
        else:
            rendered.append(f"{prefix}{connector}{name}")
    return rendered


def _candidate_keys(explicit_key: str | None, fallback_keys: tuple[str, ...]) -> tuple[str, ...]:
    if explicit_key is None:
        return fallback_keys
    return (explicit_key, *fallback_keys)


def _merge_tool_projections(
    projections: Iterable[ToolProjection | None],
) -> ToolProjection | None:
    merged_content: (
        list[ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent] | None
    ) = None
    merged_locations: list[ToolCallLocation] | None = None
    merged_status: ToolCallStatus | None = None
    merged_title: str | None = None
    saw_projection = False
    for projection in projections:
        if not isinstance(projection, ToolProjection):
            continue
        saw_projection = True
        if projection.content is not None:
            merged_content = projection.content
        if projection.locations is not None:
            merged_locations = projection.locations
        if projection.status is not None:
            merged_status = projection.status
        if projection.title is not None:
            merged_title = projection.title
    if not saw_projection:
        return None
    return ToolProjection(
        content=merged_content,
        locations=merged_locations,
        status=merged_status,
        title=merged_title,
    )


def _preserve_file_diff_content(
    *,
    known_start: ToolCallStart | None,
    projection: ToolProjection | None,
) -> list[ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent] | None:
    if projection is None or projection.content is None:
        return None
    if known_start is None or known_start.content is None:
        return projection.content
    if len(projection.content) != len(known_start.content):
        return projection.content
    if not projection.content or not known_start.content:
        return projection.content
    if not all(isinstance(content, FileEditToolCallContent) for content in projection.content):
        return projection.content
    if not all(isinstance(content, FileEditToolCallContent) for content in known_start.content):
        return projection.content
    return known_start.content


def _stringify_value(value: Any, serialized_fallback: str | None) -> str:
    if isinstance(value, str):
        return value
    if serialized_fallback is not None:
        return serialized_fallback
    return str(value)


def _read_existing_text(path: str, *, cwd: Path | None) -> str:
    target_path = Path(path)
    if not target_path.is_absolute():
        if cwd is None:
            return ""
        target_path = cwd / target_path
    try:
        if not target_path.exists() or not target_path.is_file():
            return ""
        return target_path.read_text(encoding="utf-8", errors="replace")
    except OSError:
        return ""


def _preview_text(text: str) -> str:
    return truncate_text(text, limit=_MAX_COMMAND_PREVIEW_CHARS)


def _single_line_preview(text: str, *, limit: int) -> str:
    return single_line_summary(text, limit=limit)


def _format_command_title(command: str) -> str:
    return f"Execute {_single_line_preview(command, limit=_MAX_COMMAND_TITLE_CHARS)}"


def _format_command_preview(command: str) -> str:
    return format_code_block(
        command,
        language="bash",
        limit=_MAX_COMMAND_PREVIEW_CHARS,
    )


def _format_harness_read_progress(path: str, output_text: str) -> str:
    preview = format_code_block(
        output_text,
        language="text",
        limit=_MAX_FILE_READ_PREVIEW_CHARS,
    )
    return f"Read `{path}`:\n\n{preview}"


def _bash_status(raw_output: Any = None, *, fallback: ToolCallStatus) -> ToolCallStatus:
    if not _is_string_keyed_object_dict(raw_output):
        return fallback
    timed_out = raw_output.get("timed_out")
    if isinstance(timed_out, bool) and timed_out:
        return "failed"
    if isinstance(timed_out, int) and timed_out != 0:
        return "failed"
    returncode = raw_output.get("returncode")
    if isinstance(returncode, int) and returncode != 0:
        return "failed"
    return fallback


def _bash_progress_content(
    *,
    raw_input: Any,
    raw_output: Any,
    serialized_output: str,
) -> list[ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent]:
    command = None
    if _is_string_keyed_object_dict(raw_input):
        for key in _COMMAND_KEYS:
            value = raw_input.get(key)
            if isinstance(value, str) and value:
                command = value
                break

    if not _is_string_keyed_object_dict(raw_output):
        return [ContentToolCallContent(type="content", content=_text_block(serialized_output))]

    status_label = (
        "failed" if _bash_status(raw_output, fallback="completed") == "failed" else "success"
    )
    sections: list[str] = [f"Status: {status_label}"]
    if command is not None:
        sections.extend(("", _format_command_preview(command)))
    returncode = raw_output.get("returncode")
    if isinstance(returncode, int):
        sections.append(f"Exit code: {returncode}")
    stdout = raw_output.get("stdout")
    if isinstance(stdout, str) and stdout:
        sections.extend(
            (
                "",
                "Stdout:",
                format_code_block(stdout, language="text", limit=_MAX_COMMAND_PREVIEW_CHARS),
            ),
        )
    stderr = raw_output.get("stderr")
    if isinstance(stderr, str) and stderr:
        sections.extend(
            (
                "",
                "Stderr:",
                format_code_block(stderr, language="text", limit=_MAX_COMMAND_PREVIEW_CHARS),
            ),
        )
    return [ContentToolCallContent(type="content", content=_text_block("\n".join(sections)))]


def _web_search_query(raw_input: dict[str, Any]) -> str | None:
    for key in ("query", "q", "search_query"):
        value = raw_input.get(key)
        if isinstance(value, str) and value:
            return value
    return None


def _web_fetch_url(raw_input: dict[str, Any]) -> str | None:
    for key in ("url", "href", "uri"):
        value = raw_input.get(key)
        if isinstance(value, str) and value:
            return value
    return None


def _format_web_search_start(raw_input: dict[str, Any]) -> str:
    lines: list[str] = []
    query = _web_search_query(raw_input)
    if query is not None:
        lines.append(f"Query: {query}")
    _append_string_list_line(lines, "Allowed domains", raw_input.get("allowed_domains"))
    _append_string_list_line(lines, "Blocked domains", raw_input.get("blocked_domains"))
    context_size = raw_input.get("search_context_size")
    if isinstance(context_size, str) and context_size:
        lines.append(f"Context size: {context_size}")
    user_location = raw_input.get("user_location")
    if _is_string_keyed_object_dict(user_location):
        location_parts = [
            str(user_location[key])
            for key in ("city", "region", "country", "timezone")
            if key in user_location and user_location[key] not in (None, "")
        ]
        if location_parts:
            lines.append(f"User location: {', '.join(location_parts)}")
    return "\n".join(lines) if lines else "Searching the web."


def _format_web_fetch_start(raw_input: dict[str, Any]) -> str:
    lines: list[str] = []
    url = _web_fetch_url(raw_input)
    if url is not None:
        lines.append(f"URL: {url}")
    _append_string_list_line(lines, "Allowed domains", raw_input.get("allowed_domains"))
    _append_string_list_line(lines, "Blocked domains", raw_input.get("blocked_domains"))
    max_content_tokens = raw_input.get("max_content_tokens")
    if isinstance(max_content_tokens, int):
        lines.append(f"Max content tokens: {max_content_tokens}")
    enable_citations = raw_input.get("enable_citations")
    if isinstance(enable_citations, bool):
        lines.append(f"Citations enabled: {'yes' if enable_citations else 'no'}")
    return "\n".join(lines) if lines else "Fetching web content."


def _append_string_list_line(lines: list[str], label: str, value: Any) -> None:
    if not isinstance(value, list) or not value:
        return
    entries = [entry for entry in value if isinstance(entry, str) and entry]
    if entries:
        lines.append(f"{label}: {', '.join(entries)}")


def _format_web_search_progress(raw_output: Any, serialized_output: str) -> str:
    results = _extract_search_results(raw_output)
    if not results:
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)
    sections: list[str] = []
    for index, result in enumerate(results[:5], start=1):
        title = _first_string(result, "title", "name") or f"Result {index}"
        url = _first_string(result, "url", "href", "link")
        snippet = _first_string(result, "body", "snippet", "description", "content")
        sections.append(f"{index}. {title}")
        if url is not None:
            sections.append(url)
        if snippet is not None:
            sections.append(_single_line_preview(snippet, limit=220))
        sections.append("")
    return "\n".join(sections).rstrip()


def _extract_search_results(raw_output: Any) -> list[dict[str, Any]] | None:
    if isinstance(raw_output, list) and all(
        _is_string_keyed_object_dict(item) for item in raw_output
    ):
        results: list[dict[str, Any]] = []
        results.extend(item for item in raw_output if _is_string_keyed_object_dict(item))
        return results
    if not _is_string_keyed_object_dict(raw_output):
        return None
    for key in ("results", "items", "data"):
        value = raw_output.get(key)
        if isinstance(value, list) and all(_is_string_keyed_object_dict(item) for item in value):
            nested_results: list[dict[str, Any]] = []
            nested_results.extend(item for item in value if _is_string_keyed_object_dict(item))
            return nested_results
    return None


def _format_web_fetch_progress(raw_output: Any, serialized_output: str) -> str:
    if _is_binary_like_content(raw_output):
        media_type = getattr(raw_output, "media_type", None)
        if isinstance(media_type, str) and media_type:
            return f"Fetched binary content ({media_type})."
        return "Fetched binary content."
    if not _is_string_keyed_object_dict(raw_output):
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)

    lines: list[str] = []
    url = _first_string(raw_output, "url", "href")
    if url is not None:
        lines.append(f"URL: {url}")
    title = _first_string(raw_output, "title", "name")
    if title is not None:
        lines.append(f"Title: {title}")
    content = _first_string(raw_output, "content", "body", "text")
    if content is not None:
        lines.extend(
            (
                "",
                "Preview:",
                format_code_block(content, language="text", limit=_MAX_WEB_PREVIEW_CHARS),
            ),
        )
    if not lines:
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)
    return "\n".join(lines)


def _format_image_generation_start(raw_input: dict[str, Any]) -> str:
    lines: list[str] = []
    prompt = _first_string(raw_input, "prompt", "input", "text")
    if prompt is not None:
        lines.append(f"Prompt: {prompt}")
    for label, key in (
        ("Quality", "quality"),
        ("Size", "size"),
        ("Aspect ratio", "aspect_ratio"),
        ("Output format", "output_format"),
        ("Background", "background"),
    ):
        value = raw_input.get(key)
        if isinstance(value, str) and value:
            lines.append(f"{label}: {value}")
    return "\n".join(lines) if lines else "Generating image."


def _format_image_generation_progress(raw_output: Any, serialized_output: str) -> str:
    if not _is_string_keyed_object_dict(raw_output):
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)
    lines: list[str] = []
    status = _first_string(raw_output, "status")
    if status is not None:
        lines.append(f"Status: {status}")
    revised_prompt = _first_string(raw_output, "revised_prompt")
    if revised_prompt is not None:
        lines.append(f"Revised prompt: {revised_prompt}")
    for label, key in (
        ("Quality", "quality"),
        ("Size", "size"),
        ("Background", "background"),
    ):
        value = raw_output.get(key)
        if isinstance(value, str) and value:
            lines.append(f"{label}: {value}")
    return "\n".join(lines) if lines else "Image generation completed."


def _format_mcp_title(tool_name: str, raw_input: dict[str, Any]) -> str:
    server_label = tool_name.split(":", 1)[1] if ":" in tool_name else tool_name
    action = _first_string(raw_input, "action")
    actual_tool_name = _first_string(raw_input, "tool_name")
    if action == "call_tool" and actual_tool_name is not None:
        return f"Call {actual_tool_name} via MCP {server_label}"
    if action == "list_tools":
        return f"List tools from MCP {server_label}"
    return f"Use MCP {server_label}"


def _format_mcp_start(tool_name: str, raw_input: dict[str, Any]) -> str:
    server_label = tool_name.split(":", 1)[1] if ":" in tool_name else tool_name
    lines = [f"Server: {server_label}"]
    action = _first_string(raw_input, "action")
    if action is not None:
        lines.append(f"Action: {action}")
    actual_tool_name = _first_string(raw_input, "tool_name")
    if actual_tool_name is not None:
        lines.append(f"Tool: {actual_tool_name}")
    tool_args = raw_input.get("tool_args")
    if _is_string_keyed_object_dict(tool_args) and tool_args:
        lines.extend(
            (
                "",
                "Arguments:",
                format_code_block(
                    _json_preview(tool_args),
                    language="json",
                    limit=_MAX_WEB_PREVIEW_CHARS,
                ),
            ),
        )
    return "\n".join(lines)


def _format_mcp_progress(raw_output: Any, serialized_output: str) -> str:
    if not _is_string_keyed_object_dict(raw_output):
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)
    lines: list[str] = []
    error = raw_output.get("error")
    if error not in (None, ""):
        lines.append(f"Error: {error}")
    tools = raw_output.get("tools")
    if isinstance(tools, list):
        lines.append(f"Tools listed: {len(tools)}")
        preview_names: list[str] = []
        for item in tools[:5]:
            if not _is_string_keyed_object_dict(item):
                continue
            name = item.get("name")
            if isinstance(name, str):
                preview_names.append(name)
        if preview_names:
            lines.append(f"Preview: {', '.join(preview_names)}")
    output = raw_output.get("output")
    if output not in (None, "", []):
        lines.extend(
            (
                "",
                "Output:",
                format_code_block(
                    _json_preview(output),
                    language="json",
                    limit=_MAX_WEB_PREVIEW_CHARS,
                ),
            ),
        )
    if not lines:
        return truncate_text(serialized_output, limit=_MAX_WEB_PREVIEW_CHARS)
    return "\n".join(lines)


def _first_string(value: dict[str, Any], *keys: str) -> str | None:
    for key in keys:
        item = value.get(key)
        if isinstance(item, str) and item:
            return item
    return None


def _json_preview(value: Any) -> str:
    try:
        return json.dumps(value, indent=2, sort_keys=True)
    except TypeError:
        return str(value)


def _is_binary_like_content(value: Any) -> bool:
    if value is None:
        return False
    media_type = getattr(value, "media_type", None)
    data = getattr(value, "data", None)
    return isinstance(media_type, str) and isinstance(data, bytes)


def _text_block(text: str) -> TextContentBlock:
    return TextContentBlock(type="text", text=text)


def _is_tool_call_part(value: Any) -> TypeIs[ToolCallPart | NativeToolCallPart]:
    return isinstance(value, (ToolCallPart, NativeToolCallPart))


def _is_tool_return_part(value: Any) -> TypeIs[ToolReturnPart | NativeToolReturnPart]:
    return isinstance(value, (ToolReturnPart, NativeToolReturnPart))


def _build_tool_start_projection(
    part: ToolCallPart | NativeToolCallPart,
    *,
    cwd: Path | None,
    projection_map: ProjectionMap | None,
) -> ToolProjection | None:
    if projection_map is None:
        return None
    return projection_map.project_start(
        part.tool_name,
        cwd=cwd,
        raw_input=part.args_as_dict(),
    )


def _build_tool_title(
    *,
    part_tool_name: str | None,
    known_start: ToolCallStart | None,
    projection: ToolProjection | None,
) -> str:
    if projection is not None and projection.title is not None:
        return projection.title
    if known_start is not None:
        return known_start.title
    return part_tool_name or ""


def _build_tool_locations(
    *,
    known_start: ToolCallStart | None,
    known_start_raw_input: dict[str, Any] | None,
    projection: ToolProjection | None,
) -> list[ToolCallLocation] | None:
    if projection is not None and projection.locations is not None:
        return projection.locations
    if known_start is not None and known_start.locations is not None:
        return known_start.locations
    if known_start_raw_input is None:
        return None
    return extract_tool_call_locations(known_start_raw_input)


def build_tool_start_update(
    part: ToolCallPart | NativeToolCallPart,
    *,
    classifier: ToolClassifier,
    cwd: Path | None = None,
    projection_map: ProjectionMap | None,
) -> ToolCallStart:
    raw_input = part.args_as_dict()
    projection = _build_tool_start_projection(
        part,
        cwd=cwd,
        projection_map=projection_map,
    )
    return ToolCallStart(
        session_update="tool_call",
        tool_call_id=part.tool_call_id,
        title=(
            projection.title
            if projection is not None and projection.title is not None
            else part.tool_name
        ),
        kind=classifier.classify(part.tool_name, raw_input),
        status="in_progress",
        content=projection.content if projection is not None else None,
        locations=(
            projection.locations
            if projection is not None and projection.locations is not None
            else extract_tool_call_locations(raw_input)
        ),
        raw_input=raw_input,
    )


def build_tool_progress_update(
    part: ToolReturnPart | NativeToolReturnPart | RetryPromptPart,
    *,
    classifier: ToolClassifier,
    cwd: Path | None = None,
    known_start: ToolCallStart | None,
    projection_map: ProjectionMap | None,
    serializer: OutputSerializer,
) -> ToolCallProgress:
    known_start_raw_input = (
        known_start.raw_input
        if known_start is not None and _is_string_keyed_object_dict(known_start.raw_input)
        else None
    )
    projection = None
    serialized_output: str | None = None
    status: ToolCallStatus | None = None
    if _is_tool_return_part(part):
        serialized_output = serializer.serialize(part.content)
        status = "completed"
        if isinstance(part, ToolReturnPart):
            status = "completed" if part.outcome == "success" else "failed"
        if projection_map is not None:
            projection = projection_map.project_progress(
                part.tool_name,
                cwd=cwd,
                raw_input=known_start_raw_input,
                raw_output=part.content,
                serialized_output=serialized_output,
                status=status,
            )
    title = _build_tool_title(
        part_tool_name=part.tool_name,
        known_start=known_start,
        projection=projection,
    )
    tool_name = part.tool_name if part.tool_name is not None else title
    kind: ToolKind = (
        known_start.kind
        if known_start is not None and known_start.kind is not None
        else classifier.classify(tool_name)
    )
    preserved_content = _preserve_file_diff_content(
        known_start=known_start,
        projection=projection,
    )
    if _is_tool_return_part(part):
        return ToolCallProgress(
            session_update="tool_call_update",
            tool_call_id=part.tool_call_id,
            title=title,
            kind=kind,
            status=(
                projection.status
                if projection is not None and projection.status is not None
                else status
            ),
            content=(
                preserved_content
                if preserved_content is not None
                else (known_start.content if known_start is not None else None)
            ),
            locations=_build_tool_locations(
                known_start=known_start,
                known_start_raw_input=known_start_raw_input,
                projection=projection,
            ),
            raw_output=serialized_output,
        )

    return ToolCallProgress(
        session_update="tool_call_update",
        tool_call_id=part.tool_call_id,
        title=title,
        kind=kind,
        status="failed",
        locations=known_start.locations if known_start is not None else None,
        raw_output=part.model_response(),
    )


def _build_progress_updates_for_message(
    message: ModelMessage,
    *,
    classifier: ToolClassifier,
    cwd: Path | None,
    known_call_starts: dict[str, ToolCallStart],
    projection_map: ProjectionMap | None,
    serializer: OutputSerializer,
) -> list[ToolCallProgress | ToolCallStart]:
    updates: list[ToolCallProgress | ToolCallStart] = []
    if isinstance(message, ModelResponse):
        for part in message.parts:
            if _is_tool_call_part(part):
                if _is_output_tool(part.tool_name):
                    continue
                if part.tool_call_id in known_call_starts:
                    continue
                start_update = build_tool_start_update(
                    part,
                    classifier=classifier,
                    cwd=cwd,
                    projection_map=projection_map,
                )
                known_call_starts[part.tool_call_id] = start_update
                updates.append(start_update)
                continue
            if isinstance(part, NativeToolReturnPart):
                if _is_output_tool(part.tool_name):
                    continue
                updates.append(
                    build_tool_progress_update(
                        part,
                        classifier=classifier,
                        cwd=cwd,
                        known_start=known_call_starts.get(part.tool_call_id),
                        projection_map=projection_map,
                        serializer=serializer,
                    ),
                )
        return updates

    for part in message.parts:
        if _is_tool_return_part(part):
            if _is_output_tool(part.tool_name):
                continue
            updates.append(
                build_tool_progress_update(
                    part,
                    classifier=classifier,
                    cwd=cwd,
                    known_start=known_call_starts.get(part.tool_call_id),
                    projection_map=projection_map,
                    serializer=serializer,
                ),
            )
        elif (
            isinstance(part, RetryPromptPart)
            and part.tool_name is not None
            and not _is_output_tool(part.tool_name)
        ):
            updates.append(
                build_tool_progress_update(
                    part,
                    classifier=classifier,
                    cwd=cwd,
                    known_start=known_call_starts.get(part.tool_call_id),
                    projection_map=projection_map,
                    serializer=serializer,
                ),
            )
    return updates


def build_tool_updates(
    messages: list[ModelMessage],
    *,
    classifier: ToolClassifier,
    cwd: Path | None = None,
    known_starts: dict[str, ToolCallStart] | None = None,
    projection_map: ProjectionMap | None,
    serializer: OutputSerializer,
) -> list[ToolCallProgress | ToolCallStart]:
    known_call_starts = dict(known_starts or {})
    updates: list[ToolCallProgress | ToolCallStart] = []
    for message in messages:
        updates.extend(
            _build_progress_updates_for_message(
                message,
                classifier=classifier,
                cwd=cwd,
                known_call_starts=known_call_starts,
                projection_map=projection_map,
                serializer=serializer,
            ),
        )
    return updates


def build_compaction_updates(
    messages: list[ModelMessage],
    *,
    known_starts: dict[str, ToolCallStart] | None = None,
    skip_providers: frozenset[str] = frozenset(),
) -> list[ToolCallProgress | ToolCallStart]:
    known_call_starts = dict(known_starts or {})
    updates: list[ToolCallProgress | ToolCallStart] = []
    created_count = 0
    for message in messages:
        if not isinstance(message, ModelResponse):
            continue
        for part in message.parts:
            if not isinstance(part, CompactionPart):
                continue
            provider_name = (part.provider_name or "unknown").lower()
            if provider_name in skip_providers:
                continue
            tool_call_id = _compaction_tool_call_id(
                part,
                provider_name=provider_name,
                known_starts=known_call_starts,
                created_count=created_count,
            )
            created_count += 1
            known_start = known_call_starts.get(tool_call_id)
            if known_start is None:
                known_start = ToolCallStart(
                    session_update="tool_call",
                    tool_call_id=tool_call_id,
                    title="Context Compaction",
                    kind="execute",
                    status="in_progress",
                    raw_input=_compaction_raw_input(part, provider_name=provider_name),
                )
                known_call_starts[tool_call_id] = known_start
                updates.append(known_start)
            updates.append(
                ToolCallProgress(
                    session_update="tool_call_update",
                    tool_call_id=tool_call_id,
                    title=known_start.title,
                    kind=known_start.kind,
                    status="completed",
                    raw_output=_format_compaction_progress(part, provider_name=provider_name),
                ),
            )
    return updates


def _compaction_tool_call_id(
    part: CompactionPart,
    *,
    provider_name: str,
    known_starts: dict[str, ToolCallStart],
    created_count: int,
) -> str:
    if part.id:
        return f"compaction:{provider_name}:{part.id}"
    candidate = f"compaction:{provider_name}:{len(known_starts) + created_count + 1}"
    while candidate in known_starts:
        created_count += 1
        candidate = f"compaction:{provider_name}:{len(known_starts) + created_count + 1}"
    return candidate


def _compaction_raw_input(
    part: CompactionPart,
    *,
    provider_name: str,
) -> dict[str, JsonValue]:
    raw_input: dict[str, JsonValue] = {"provider": provider_name}
    if part.id:
        raw_input["compaction_id"] = part.id
    return raw_input


def _format_compaction_progress(
    part: CompactionPart,
    *,
    provider_name: str,
) -> str:
    lines = [
        f"Provider: {provider_name}",
        "Status: history compacted",
    ]
    if part.content:
        lines.extend(("", "Summary:", truncate_text(part.content, limit=_MAX_WEB_PREVIEW_CHARS)))
    elif part.provider_details:
        lines.append("Compaction payload stored for round-trip.")
    else:
        lines.append("Compaction completed.")
    if part.id:
        lines.append(f"Compaction id: {part.id}")
    return "\n".join(lines)



from __future__ import annotations as _annotations

from dataclasses import dataclass, field

from acp.schema import (
    ContentToolCallContent,
    TextContentBlock,
    ToolCallProgress,
    ToolCallStart,
    ToolCallStatus,
)

__all__ = ("HookEvent", "HookProjectionMap")


@dataclass(frozen=True, slots=True, kw_only=True)
class HookEvent:
    event_id: str
    hook_name: str
    summary: str
    detail: str | None = None
    status: ToolCallStatus | None = None
    hidden: bool = False


@dataclass(slots=True, kw_only=True)
class HookProjectionMap:
    title_prefix: str = "Hook"
    hidden_event_ids: set[str] = field(default_factory=set)

    def build_start_update(
        self,
        *,
        tool_call_id: str,
        event: HookEvent,
    ) -> ToolCallStart | None:
        if event.hidden:
            self.hidden_event_ids.add(event.event_id)
            return None
        return ToolCallStart(
            session_update="tool_call",
            tool_call_id=tool_call_id,
            title=f"{self.title_prefix}: {event.hook_name}",
            kind="execute",
            status="in_progress",
            raw_input={"event_id": event.event_id, "summary": event.summary},
            content=[
                ContentToolCallContent(
                    type="content",
                    content=TextContentBlock(type="text", text=event.summary),
                ),
            ],
        )

    def build_progress_update(
        self,
        *,
        tool_call_id: str,
        event: HookEvent,
    ) -> ToolCallProgress | None:
        if event.hidden:
            self.hidden_event_ids.add(event.event_id)
            return None
        detail = event.detail or event.summary
        return ToolCallProgress(
            session_update="tool_call_update",
            tool_call_id=tool_call_id,
            title=f"{self.title_prefix}: {event.hook_name}",
            kind="execute",
            status=event.status or "completed",
            raw_output=detail,
            content=[
                ContentToolCallContent(
                    type="content",
                    content=TextContentBlock(type="text", text=detail),
                ),
            ],
        )



from __future__ import annotations as _annotations

from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from contextvars import ContextVar, Token
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Protocol, TypeAlias

from acp.schema import PlanEntry
from pydantic import BaseModel

if TYPE_CHECKING:
    from .session.state import AcpSessionContext

PlanGenerationType: TypeAlias = Literal["tools", "structured"]

__all__ = (
    "NativePlanGeneration",
    "PlanGenerationType",
    "TaskPlan",
    "acp_get_plan",
    "acp_mark_plan_done",
    "acp_set_plan",
    "acp_update_plan_entry",
    "native_plan_tools",
)


class TaskPlan(BaseModel):
    plan_md: str
    plan_entries: list[PlanEntry]


NativePlanGeneration = TaskPlan


class _ActivePlanRuntime(Protocol):
    def format_native_plan(self, session: AcpSessionContext) -> str: ...

    async def persist_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        entries: Sequence[PlanEntry],
        plan_markdown: str | None,
    ) -> None: ...

    async def update_native_plan_entry(
        self,
        session: AcpSessionContext,
        *,
        index: int,
        status: str | None = None,
        content: str | None = None,
        priority: str | None = None,
    ) -> PlanEntry: ...


@dataclass(slots=True, frozen=True)
class _BoundPlanContext:
    runtime: _ActivePlanRuntime
    session: AcpSessionContext


_ACTIVE_PLAN_CONTEXT: ContextVar[_BoundPlanContext | None] = ContextVar(
    "langchain_acp_active_plan_context",
    default=None,
)


@contextmanager
def _bind_native_plan_context(
    runtime: _ActivePlanRuntime,
    session: AcpSessionContext,
) -> Iterator[None]:
    token: Token[_BoundPlanContext | None] = _ACTIVE_PLAN_CONTEXT.set(
        _BoundPlanContext(runtime=runtime, session=session),
    )
    try:
        yield
    finally:
        _ACTIVE_PLAN_CONTEXT.reset(token)


def _active_plan_context() -> _BoundPlanContext | None:
    return _ACTIVE_PLAN_CONTEXT.get()


def acp_get_plan() -> str:
    """Return the saved plan and numbered entries."""
    context = _active_plan_context()
    if context is None:
        return "No active ACP session is bound."
    return context.runtime.format_native_plan(context.session)


async def acp_set_plan(entries: list[PlanEntry], plan_md: str | None = None) -> str:
    """Replace the current ACP-owned plan state."""
    context = _active_plan_context()
    if context is None:
        return "No active ACP session is bound."
    await context.runtime.persist_native_plan_state(
        context.session,
        entries=entries,
        plan_markdown=plan_md,
    )
    return f"Recorded {len(entries)} plan entries."


async def acp_update_plan_entry(
    index: int,
    status: str | None = None,
    content: str | None = None,
    priority: str | None = None,
) -> str:
    """Update a single plan entry by its 1-based index."""
    context = _active_plan_context()
    if context is None:
        return "No active ACP session is bound."
    updated_entry = await context.runtime.update_native_plan_entry(
        context.session,
        index=index,
        status=status,
        content=content,
        priority=priority,
    )
    return (
        f"Updated plan entry {index}: "
        f"[{updated_entry.status}] ({updated_entry.priority}) {updated_entry.content}"
    )


async def acp_mark_plan_done(index: int) -> str:
    """Mark a single plan entry completed by its 1-based index."""
    context = _active_plan_context()
    if context is None:
        return "No active ACP session is bound."
    updated_entry = await context.runtime.update_native_plan_entry(
        context.session,
        index=index,
        status="completed",
    )
    return f"Marked plan entry {index} as completed: {updated_entry.content}"


def native_plan_tools() -> tuple[object, ...]:
    return (
        acp_get_plan,
        acp_set_plan,
        acp_update_plan_entry,
        acp_mark_plan_done,
    )



from __future__ import annotations as _annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, cast

import pytest
from acpremote import cli as cli_module
from acpremote.command import CommandOptions


@dataclass(slots=True)
class _FakeSocket:
    port: int = 4321

    def getsockname(self) -> tuple[str, int]:
        return ("127.0.0.1", self.port)


@dataclass(slots=True)
class _FakeNonTcpSocket:
    def getsockname(self) -> str:
        return "pipe"


@dataclass(slots=True)
class _FakeServer:
    sockets: list[_FakeSocket] = field(default_factory=lambda: [_FakeSocket()])
    serve_forever_calls: int = 0
    close_calls: int = 0
    wait_closed_calls: int = 0

    async def serve_forever(self) -> None:
        self.serve_forever_calls += 1

    def close(self) -> None:
        self.close_calls += 1

    async def wait_closed(self) -> None:
        self.wait_closed_calls += 1


def _write_native_agent_module(tmp_path: Path) -> None:
    (tmp_path / "native_cli_app.py").write_text(
        "\n".join(
            (
                "from __future__ import annotations as _annotations",
                "",
                "from typing import Any",
                "",
                "class NativeAgent:",
                "    async def initialize(self, *args: Any, **kwargs: Any) -> None:",
                "        return None",
                "    async def new_session(self, *args: Any, **kwargs: Any) -> None:",
                "        return None",
                "    async def prompt(self, *args: Any, **kwargs: Any) -> None:",
                "        return None",
                "    async def cancel(self, *args: Any, **kwargs: Any) -> None:",
                "        return None",
                "    def on_connect(self, conn: Any) -> None:",
                "        return None",
                "",
                "ignored = object()",
                "agent = NativeAgent()",
            ),
        )
        + "\n",
        encoding="utf-8",
    )


def test_acpremote_expose_command_builds_stdio_server(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    server = _FakeServer()
    calls: list[dict[str, Any]] = []
    monkeypatch.setenv("REMOTE_TOKEN", "secret")

    async def fake_serve_stdio_command(
        command_options: CommandOptions,
        **kwargs: Any,
    ) -> _FakeServer:
        calls.append({"command_options": command_options, "kwargs": kwargs})
        return server

    monkeypatch.setattr(cli_module, "serve_stdio_command", fake_serve_stdio_command)

    exit_code = cli_module.main(
        [
            "expose",
            "--host",
            "0.0.0.0",
            "--port",
            "9090",
            "--mount-path",
            "remote",
            "--token-env",
            "REMOTE_TOKEN",
            "--cwd",
            str(tmp_path),
            "--env",
            "A=1",
            "--env",
            "EMPTY=",
            "--stderr-mode",
            "discard",
            "--terminate-timeout",
            "1.5",
            "--",
            "python",
            "agent.py",
            "--flag",
        ],
    )

    assert exit_code == 0
    assert server.serve_forever_calls == 1
    assert server.close_calls == 1
    assert server.wait_closed_calls == 1
    command_options = cast("CommandOptions", calls[0]["command_options"])
    assert command_options.command == ("python", "agent.py", "--flag")
    assert command_options.cwd == str(tmp_path)
    assert command_options.env == {"A": "1", "EMPTY": ""}
    assert command_options.stderr_mode == "discard"
    assert command_options.terminate_timeout == 1.5
    assert calls[0]["kwargs"] == {
        "host": "0.0.0.0",
        "port": 9090,
        "mount_path": "remote",
        "bearer_token": "secret",
    }
    assert "ws://0.0.0.0:4321/remote/ws" in capsys.readouterr().err


def test_acpremote_serve_loads_native_agent_target(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    _write_native_agent_module(tmp_path)
    server = _FakeServer()
    calls: list[dict[str, Any]] = []

    async def fake_serve_acp(agent: Any, **kwargs: Any) -> _FakeServer:
        calls.append({"agent": agent, "kwargs": kwargs})
        return server

    monkeypatch.setattr(cli_module, "serve_acp", fake_serve_acp)

    exit_code = cli_module.main(
        [
            "serve",
            "native_cli_app:agent",
            "--path",
            str(tmp_path),
            "--host",
            "127.0.0.1",
            "--port",
            "7070",
            "--mount-path",
            "/acp",
            "--bearer-token",
            "direct-token",
        ],
    )

    assert exit_code == 0
    assert calls[0]["agent"].__class__.__name__ == "NativeAgent"
    assert calls[0]["kwargs"] == {
        "host": "127.0.0.1",
        "port": 7070,
        "mount_path": "/acp",
        "bearer_token": "direct-token",
    }
    assert server.close_calls == 1


def test_acpremote_serve_uses_latest_native_agent_when_attribute_is_omitted(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    _write_native_agent_module(tmp_path)
    calls: list[Any] = []

    async def fake_serve_acp(agent: Any, **kwargs: Any) -> _FakeServer:
        del kwargs
        calls.append(agent)
        return _FakeServer()

    monkeypatch.setattr(cli_module, "serve_acp", fake_serve_acp)

    exit_code = cli_module.main(["serve", "native_cli_app", "--path", str(tmp_path)])

    assert exit_code == 0
    assert calls[0].__class__.__name__ == "NativeAgent"


def test_acpremote_mirror_runs_remote_proxy(monkeypatch: pytest.MonkeyPatch) -> None:
    proxy_agent = object()
    calls: list[Any] = []

    def fake_connect_acp(addr: str, *, bearer_token: str | None = None) -> object:
        calls.append(("connect", addr, bearer_token))
        return proxy_agent

    async def fake_run_agent(agent: object) -> None:
        calls.append(("run", agent))

    monkeypatch.setattr(cli_module, "connect_acp", fake_connect_acp)
    monkeypatch.setattr(cli_module, "run_agent", fake_run_agent)

    exit_code = cli_module.main(["mirror", "ws://example.test/acp/ws", "--bearer-token", "token"])

    assert exit_code == 0
    assert calls == [
        ("connect", "ws://example.test/acp/ws", "token"),
        ("run", proxy_agent),
    ]


def test_acpremote_cli_reports_user_errors(capsys: pytest.CaptureFixture[str]) -> None:
    assert cli_module.main(["expose"]) == 2
    assert "Command must not be empty." in capsys.readouterr().err

    assert cli_module.main(["expose", "--env", "INVALID", "python"]) == 2
    assert "Environment overrides must use `KEY=VALUE`." in capsys.readouterr().err

    assert cli_module.main(["mirror", "ws://example.test/acp/ws", "--token-env", "MISSING"]) == 2
    assert "`MISSING` is not set or is empty." in capsys.readouterr().err


def test_acpremote_serve_rejects_non_native_target(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    (tmp_path / "not_native.py").write_text("agent = object()\n", encoding="utf-8")

    exit_code = cli_module.main(["serve", "not_native:agent", "--path", str(tmp_path)])

    assert exit_code == 2
    assert "Target must resolve to a native `acp.interfaces.Agent`" in capsys.readouterr().err


def test_acpremote_cli_returns_130_on_keyboard_interrupt(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def interrupting_connect_acp(addr: str, *, bearer_token: str | None = None) -> object:
        del addr, bearer_token
        raise KeyboardInterrupt

    monkeypatch.setattr(cli_module, "connect_acp", interrupting_connect_acp)

    assert cli_module.main(["mirror", "ws://example.test/acp/ws"]) == 130


def test_acpremote_cli_banner_handles_servers_without_tcp_port(
    capsys: pytest.CaptureFixture[str],
) -> None:
    server_without_sockets = _FakeServer(sockets=[])
    server_with_non_tcp_socket = _FakeServer(sockets=[cast("Any", _FakeNonTcpSocket())])

    cli_module._print_server_banner(server_without_sockets, host="127.0.0.1", mount_path="/acp")

    assert "ws://127.0.0.1/acp/ws" in capsys.readouterr().err
    assert cli_module._server_port(server_with_non_tcp_socket) is None


def test_acpremote_cli_private_validation_paths(tmp_path: Path) -> None:
    assert cli_module._parse_env_overrides(()) is None

    with pytest.raises(cli_module.AcpRemoteCliError, match="module name"):
        cli_module._parse_target_ref("")
    with pytest.raises(cli_module.AcpRemoteCliError, match="attribute cannot be empty"):
        cli_module._parse_target_ref("demo:")
    with pytest.raises(cli_module.AcpRemoteCliError, match="Could not import module"):
        cli_module._load_target("missing_cli_module:agent", import_roots=())

    (tmp_path / "empty_cli_app.py").write_text("value = object()\n", encoding="utf-8")

    with pytest.raises(cli_module.AcpRemoteCliError, match="defines no native ACP agent"):
        cli_module._load_target("empty_cli_app", import_roots=(str(tmp_path),))
    with pytest.raises(cli_module.AcpRemoteCliError, match="missing attribute"):
        cli_module._load_target("empty_cli_app:agent", import_roots=(str(tmp_path),))
